#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""Some base classes, stubs and types."""
#from __future__ import annotations

#__all__ = ["jit"]


########### IMPORTS ########### {{{1
from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, MutableMapping, overload
from typing_extensions import runtime, Protocol
from operator import mul, truediv
import math
import sys
import importlib
import pickle
import textwrap
import copy
from inspect import signature
from functools import partial
import warnings

from qdpy.utils import *


########### INTERFACES AND STUBS ########### {{{1
T = TypeVar("T")
DomainLike = Tuple[float, float]
ShapeLike = Tuple[int, ...]
#BackendLike = Union[MutableSet[T], MutableSequence[T]]
FitnessValuesLike = Sequence[T]
FeaturesLike = Sequence

@runtime
class FitnessLike(Protocol):
    """Fitness protocol inspired from (and compatible with) DEAP Fitness class."""
    weights: FitnessValuesLike
    def dominates(self, other: Any, obj: Any = slice(None)) -> bool: ...
    def getValues(self) -> FitnessValuesLike: ...
    def setValues(self, values: FitnessValuesLike) -> None: ...
    def delValues(self) -> None: ...
    @property
    def values(self) -> FitnessValuesLike: ...
    @values.setter
    def values(self, values: FitnessValuesLike) -> None: ...
    @values.deleter
    def values(self) -> None: ...
    @property
    def valid(self) -> bool: ...
#    @overload
#    def __getitem__(self, i: int) -> Any: ...
#    @overload
#    def __getitem__(self, s: slice) -> Sequence[Any]: ...
#    def __len__(self) -> int: ...


@runtime
class IndividualLike(Protocol):
    name: str
    fitness: FitnessLike
    features: FeaturesLike
    elapsed: float
    def dominates(self, other: Any) -> bool: ...
    def __setitem__(self, key, values) -> None: ...


#@runtime
#class DEAPFitnessLike(FitnessLike):
#    pass
#
#@runtime
#class DEAPIndividualLike(Protocol):
#    fitness: DEAPStyleFitnessLike
#    features: FeaturesLike


#FitnessLike = Sequence
FitnessGetter = Callable[[T], FitnessLike]
FeaturesGetter = Callable[[T], FeaturesLike]
GridIndexLike = ShapeLike


########### CONCURRENCY ########### {{{1

@runtime
class FutureLike(Protocol):
    def cancelled(self) -> bool: ...
    def done(self) -> bool: ...
    def result(self) -> Any: ...
    def exception(self) -> Any: ...
    def add_done_callback(self, fn: Callable, **kwargs: Any) -> None: ...
    def as_completed(self, fs: Sequence) -> Any: ...

@runtime
class ExecutorLike(Protocol):
    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> FutureLike: ...
    def map(self, fn: Callable, *iterables: Any) -> Iterable: ...
    def shutdown(self, wait: bool = True) -> None: ...


class SequentialFuture(FutureLike):
    """Future object returned by `SequentialExecutor` methods."""
    _result: Any
    def __init__(self, result: Any) -> None:
        self._result = result
    def cancelled(self) -> bool:
        return False
    def done(self) -> bool:
        return True
    def result(self) -> Any:
        return self._result
    def exception(self) -> Any:
        return None
    def add_done_callback(self, fn: Callable, **kwargs: Any) -> None:
        # Call directly the callback
        fn()

class SequentialExecutor(ExecutorLike):
    """Executor that runs sequentially (no parallelism)."""
    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> FutureLike:
        return SequentialFuture(fn(*args, **kwargs))
    def map(self, fn: Callable, *iterables: Any) -> Iterable:
        # From cpython code 'concurrent/futures/_base.py'
        fs = [self.submit(fn, *args) for args in zip(*iterables)]
        # Yield must be hidden in closure so that the futures are submitted
        # before the first iterator value is required.
        def result_iterator():
            try:
                # reverse to keep finishing order
                fs.reverse()
                while fs:
                    # Careful not to keep a reference to the popped future
                    yield fs.pop().result()
            finally:
                for future in fs:
                    future.cancel()
        return result_iterator()
    def shutdown(self, wait: bool = True) -> None:
        pass


try:
    import scoop.futures
    class ScoopExecutor(ExecutorLike):
        """Executor that encapsulate scoop concurrency functions. Need the scoop package to be installed and importable."""
        def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> FutureLike:
            assert len(kwargs) == 0, f"`ScoopExecutor.submit` does not handle **kwargs."
            return scoop.futures.submit(fn, *args)
        def map(self, fn: Callable, *iterables: Any) -> Iterable:
            return scoop.futures.map(fn, *iterables)
        def shutdown(self, wait: bool = True) -> None:
            scoop.futures.shutdown(wait)
    _scoop_as_completed: Optional[Callable] = scoop.futures.as_completed
    _scoop_future_class: Optional[Any] = scoop.futures.Future

except ImportError:
    class ScoopExecutor(ExecutorLike): # type: ignore
        """Executor that encapsulate scoop concurrency functions. Need the scoop package to be installed and importable."""
        def __init__(self) -> None:
            raise NotImplementedError("`ScoopExecutor` needs the scoop package to be installed and importable.")
        def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> FutureLike:
            raise NotImplementedError("`ScoopExecutor` needs the scoop package to be installed and importable.")
        def map(self, fn: Callable, *iterables: Any) -> Iterable:
            raise NotImplementedError("`ScoopExecutor` needs the scoop package to be installed and importable.")
        def shutdown(self, wait: bool = True) -> None:
            raise NotImplementedError("`ScoopExecutor` needs the scoop package to be installed and importable.")
    _scoop_as_completed: Optional[Callable] = None # type: ignore
    _scoop_future_class: Optional[Any] = None # type: ignore




def generic_as_completed(fs: Sequence[FutureLike], timeout: Optional[int]=None) -> FutureLike:
    """Generic `as_completed` function, that can work with any kind of FutureLike object, including
    Futures from `concurrent.futures` and `scoop.futures`.
    Assume that all items from `fs` are instance of the same `FutureLike` class."""
    if len(fs) == 0:
        raise ValueError(f"`fs` must not be empty.")
    elif len(fs) == 1:
        _ = fs[0].result()
        return fs[0]
    # Verify if the futures are instances of SequentialFuture (assume all items have the same class)
    if isinstance(fs[0], SequentialFuture):
        return fs[0]
    if _scoop_future_class is not None and isinstance(fs[0], _scoop_future_class) and _scoop_as_completed is not None:
        return next(_scoop_as_completed(fs, timeout))
    # Get the class and module of the first item (assume all items have the same class)
    futures_class = fs[0].__class__
    futures_module = fs[0].__module__
    # Verify if a `as_completed` function is already available
    m = importlib.import_module(futures_module)
    if hasattr(m, "as_completed"):
        return next(m.as_completed(fs, timeout)) # type: ignore
    else:
        raise ValueError(f"Futures from `fs` are of unknown type, with unknown `as_completed` corresponding function.")



# TODO Support for Dask
class ParallelismManager(object):
    """Simplify parallelism handling by generating on-the-fly executor and toolboxes to conform to a
    specified kind of parallelism. It can be used with qdpy QDAlgorithm classes, or with DEAP toolboxes.
    It currently handle three kind of parallelism: Sequential (no parallelism), Python PEP3148 concurrent.futures,
    and SCOOP (https://github.com/soravux/scoop)."""

    executor: ExecutorLike
    toolbox: Optional[Any]
    orig_toolbox: Optional[Any]
    max_workers: Optional[int]
    _parallelism_type: str = ""
    _parallelism_type_list: Sequence[str] = ['none', 'sequential', 'multiprocessing', 'concurrent', 'multithreading', 'scoop']

    def __init__(self, parallelism_type: str = "multiprocessing", max_workers: Optional[int] = None,
            toolbox: Optional[Any] = None) -> None:
        self.parallelism_type = parallelism_type
        self.max_workers = max_workers
        self.orig_toolbox = toolbox
        self.toolbox = copy.deepcopy(toolbox) if toolbox is not None else None

    @property
    def parallelism_type(self) -> str:
        return self._parallelism_type

    @parallelism_type.setter
    def parallelism_type(self, val: str) -> None:
        if len(self._parallelism_type) > 0:
            raise NotImplementedError("`parallelism_type` cannot be set after initialisation.")
        lval = val.lower()
        if not lval in self._parallelism_type_list:
            raise ValueError(f"`parallelism_type` can only be from [{','.join(self._parallelism_type_list)}].")
        self._parallelism_type = lval

    def __enter__(self) -> Any:
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def open(self) -> Any:
        """Open and prepare a new executor according to `self.parallelism_type`."""
        if self.parallelism_type == "none" or self.parallelism_type == "sequential":
            self.executor = SequentialExecutor()
        elif self.parallelism_type == "multiprocessing" or self.parallelism_type == "concurrent":
            import concurrent.futures
            if sys.version_info >= (3, 7):
                import multiprocessing
                ctx = multiprocessing.get_context("forkserver")
                self.executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers, mp_context=ctx) # type: ignore
            else:
                self.executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers) # type: ignore
        elif self.parallelism_type == "multithreading":
            import concurrent.futures
            self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) # type: ignore
        elif self.parallelism_type == "scoop":
            import scoop
            self.executor = ScoopExecutor()
        else:
            raise ValueError("Unknown parallelism_type: '%s'" % self.parallelism_type)
        if self.toolbox is not None:
            self.toolbox.register("map", self.executor.map)
        return self

    def close(self) -> None:
        """Close the current executor."""
        self.executor.shutdown()




########### BASE OPTIMISATION CLASSES ########### {{{1

class Fitness(FitnessLike, Sequence[Any]):
    """Fitness implementation inspired from DEAP Fitness class. It can be used without problem with most (propably all) DEAP methods."""

    weights: FitnessValuesLike = ()
    wvalues: FitnessValuesLike = ()

    def __new__(cls, values: FitnessValuesLike=(), weights: Optional[FitnessValuesLike]=None):
        return super(Fitness, cls).__new__(cls)

    def __init__(self, values: FitnessValuesLike=(), weights: Optional[FitnessValuesLike]=None) -> None:
        if weights is None:
            self.weights = tuple([-1.0 for _ in range(len(values))]) # Defaults to minimisation
        else:
            self.weights = weights
        if len(self.weights) != len(values):
            raise ValueError("``values`` and ``weights`` must have the same length.")
        self.values = values

    @property
    def values(self) -> FitnessValuesLike:
        return tuple(map(truediv, self.wvalues, self.weights))

    @values.setter
    def values(self, values: FitnessValuesLike) -> None:
        try:
            self.wvalues = tuple(map(mul, values, self.weights))
        except TypeError:
            raise ValueError("Invalid ``value`` parameter.")

    @values.deleter
    def values(self) -> None:
        self.wvalues = ()

    def getValues(self) -> FitnessValuesLike:
        return self.values

    def setValues(self, values: FitnessValuesLike) -> None:
        self.values = values

    def delValues(self) -> None:
        del self.values

    # FROM DEAP
    def dominates(self, other: Any, obj: Any = slice(None)) -> bool:
        """Return true if each objective of ``self`` is not strictly worse than
        the corresponding objective of ``other`` and at least one objective is
        strictly better.
        """
#    :param obj: Slice indicating on which objectives the domination is
#                tested. The default value is `slice(None)`, representing
#                every objectives.  """
        not_equal: bool = False
        for self_wvalue, other_wvalue in zip(self.wvalues[obj], other.wvalues[obj]):
            if self_wvalue > other_wvalue:
                not_equal = True
            elif self_wvalue < other_wvalue:
                return False
        return not_equal

    @property
    def valid(self) -> bool:
        """Assess if a fitness is valid or not."""
        return len(self.wvalues) != 0

    def __len__(self) -> int:
        return len(self.wvalues)

#    @overload
#    def __getitem__(self, index: int) -> T: ...
#
#    @overload
#    def __getitem__(self, s: slice) -> Sequence: ...

    def __getitem__(self, key):
        return self.values[key]

    def __contains__(self, key: Any) -> bool:
        return key in self.values

    def __iter__(self) -> Iterator:
        return iter(self.values)

    def __reversed__(self) -> Iterator:
        return reversed(self.values)


    def __hash__(self) -> int:
        return hash(self.wvalues)

    def __gt__(self, other: Any) -> bool:
        return not self.__le__(other)

    def __ge__(self, other: Any) -> bool:
        return not self.__lt__(other)

    def __le__(self, other: Any) -> bool:
        return self.wvalues <= other.wvalues

    def __lt__(self, other: Any) -> bool:
        return self.wvalues < other.wvalues

    def __eq__(self, other: Any) -> bool:
        return self.wvalues == other.wvalues

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return str(self.values if self.valid else tuple())

    def __repr__(self) -> str:
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))


class Individual(list, IndividualLike):
    """Qdpy Individual class. Note that containers and algorithms all use internally either the QDPYIndividualLike Protocol or the IndividualWrapper class, so you can easily declare an alternative class to Individual. TODO""" # TODO

    name: str
    fitness: FitnessLike
    features: FeaturesLike
    elapsed: float = math.nan

    def __init__(self, iterable: Optional[Iterable] = None,
            name: Optional[str] = None,
            fitness: Optional[FitnessLike] = None, features: FeaturesLike = ()) -> None:
        if iterable is not None:
            self.extend(iterable)
        self.name = name if name else ""
        self.fitness = fitness if fitness is not None else Fitness()
        self.features = features

    def __repr__(self) -> str:
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))

    def dominates(self, other: Any) -> bool:
        """Return true if ``self`` dominates ``other``. """
        return self.fitness.dominates(other.fitness)

    # TODO : improve performance ! (quick and dirty solution !)
    def __hash__(self):
        return hash(tuple(self))


## TODO
#class IndividualWrapper(Individual):
#    """TODO""" # TODO
#
#    AnyIndividualLike = Union[QDPYIndividualLike, DEAPStyleIndividualLike]
#
#    _ind: AnyIndividualLike
#    fitness: FitnessLike
#    features: FeaturesLike
#
#    def __init__(self, individual: AnyIndividualLike)



########### BASE CLASSES ########### {{{1
class Summarisable(object):
    """Describes a class that can be summarised by using the `self.summary` method.
    The summarised information is provided by the `self.__get_summary_state__` method. """


    def __get_summary_state__(self) -> Mapping[str, Any]:
        """Return a dictionary containing the relevant entries to build a summary of the class.
        By default, it includes all public attributes of the class. Must be overrode by subclasses."""
        # Find public attributes
        entries = {}
        for k, v in inspect.getmembers(self):
            if not k.startswith('_') and not inspect.ismethod(v):
                try:
                    entries[k] = v
                except Exception:
                    pass
        return entries


    def summary(self, max_depth: Optional[int] = None, max_entry_length: Optional[int] = 250) -> str:
        """Return a summary description of the class.
        The summarised information is provided by the `self.__get_summary_state__` method.

        Parameters
        ----------
        :param max_depth: Optional[int]
            The maximal recursion depth allowed. Used to summarise attributes of `self` that are also Summarisable.
            If the maximal recursion depth is reached, the attribute is only described with a reduced representation (`repr(attribute)`).
            If `max_depth` is set to None, there are no recursion limit.
        :param max_entry_length: Optional[int]
            If `max_entry_length` is not None, the description of a non-Summarisable entry exceeding `max_entry_length`
            is cropped to this limit.
        """
        res: str = f"Summary {self.__class__.__name__}:\n"
        subs_max_depth = max_depth - 1 if max_depth is not None else None
        summary_state = self.__get_summary_state__()
        for i, k in enumerate(summary_state.keys()):
            v = summary_state[k]
            res += f"  {k}:"
            if isinstance(v, Summarisable):
                if max_depth is None or max_depth > 0:
                    res += textwrap.indent(v.summary(subs_max_depth), '  ')
                else:
                    res += f" {repr(v)}"
            else:
                str_v = f" {v}"
                str_v = str_v.replace("\n", " ")
                if max_entry_length is not None and len(str_v) > max_entry_length:
                    str_v = str_v[:max_entry_length - 4] + " ..."
                res += str_v
            if i != len(summary_state) - 1:
                res += "\n"
        return res

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"



class Saveable(object):
    """Describes a class with internal information that can be saved into an output file.
    The list of attributes from `self` that are saved in the output file are provided by the method `self.__get_saved_state__`.
    """

    def __get_saved_state__(self) -> Mapping[str, Any]:
        """Return a dictionary containing the relevant information to save.
        By default, it includes all public attributes of the class. Must be overrode by subclasses."""
        # Find public attributes
        entries = {k:v for k,v in inspect.getmembers(self) if not k.startswith('_') and not inspect.ismethod(v)}
        return entries

    def save(self, output_path: str, output_type: str = "pickle") -> None:
        """Save the into an output file.

        Parameters
        ----------
        :param output_path: str
            Path of the output file.
        :param output_type: str
            Type of the output file. Currently, only supports "pickle".
        """
        if output_type != "pickle":
            raise ValueError(f"Invalid `output_type` value. Currently, only supports 'pickle'.")
        saved_state = self.__get_saved_state__()
        with open(output_path, "wb") as f:
            pickle.dump(saved_state, f)


class Copyable(object):
    """Describes a class capable to be copied and deepcopied."""

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result



class CreatableFromConfig(object):
    """Describe a class capable to be created from a configuration dictionary."""

    @classmethod
    def from_config(cls, config: Mapping[str, Any], **kwargs: Any) -> Any:
        """Create a class using the information from `config` to call the class constructor.

        Parameters
        ----------
        :param config: Mapping[str, Any]
            The configuration mapping used to create the class. For each entry, the key corresponds
            to the name of a parameter of the constructor of the class.
        :param kwargs: Any
            Additional information used to create the class. The configuration entries from kwargs
            take precedence over the entry in `config`.
        """
        final_kwargs = {**config, **kwargs}
        return cls(**final_kwargs) # type: ignore


# Inspired from Nevergrad (MIT License) Registry class (https://github.com/facebookresearch/nevergrad/blob/master/nevergrad/common/decorators.py)
class Registry(dict):
    """Registers function or classes as a dict."""

    def __init__(self) -> None:
        super().__init__()
        self._information: MutableMapping[str, Mapping] = {}

    def register(self, obj: Any, info: Optional[Mapping[Any, Any]] = None) -> Any:
        """Decorator method for registering functions/classes
        The `info` variable can be filled up using the register_with_info
        decorator instead of this one.
        """
        name = obj.__name__
        if name in self:
            #raise RuntimeError(f'Encountered a name collision "{name}".')
            warnings.warn(f'Encountered a name collision "{name}".')
            return self[name]
        self[name] = obj
        if info is not None:
            self._information[name] = info
        return obj

    def get_info(self, name: str) -> Mapping[str, Any]:
        if name not in self:
            raise ValueError(f'`{name}` is not registered.')
        return self._information.setdefault(name, {})

registry = Registry()


class Factory(dict):
    """Build objects from configuration."""

    def __init__(self, registry: Registry=registry) -> None:
        super().__init__()
        self.registry = registry


    def _get_name(self, obj: Any, key: Optional[str], config: Mapping[str, Any]) -> Optional[str]:
        """Check if `obj` possess a name, and return it.""" 
        name: str = ""
        if hasattr(obj, "name"):
            return obj.name
        elif key is not None:
            return key
        elif "name" in config:
            return config["name"]
        else:
            return None


    def _build_internal(self, config: MutableMapping[str, Any], default_params: Mapping[str, Any] = {}, **kwargs: Any) -> Any:
        # Recursively explore config and create objects for each entry possessing a "type" configuration entry
        built_objs = []
        default_config = {**default_params} 
        for k, v in config.items():
            if is_iterable(v) and "type" in v:
                sub_obj = self._build_internal(v, {**default_config, "name": k})
                sub_name = self._get_name(sub_obj, k, v)
                if sub_name is not None and len(sub_name) > 0 and not sub_name in self:
                    self[sub_name] = sub_obj
                    built_objs.append(sub_obj)
                config[k] = sub_obj
            elif isinstance(v, str) and v in self:
                config[k] = self[v]
            elif is_iterable(v):
                for i, val in enumerate(v):
                    if isinstance(val, str) and val in self:
                        config[k][i] = self[val]
            if not is_iterable(v) or (is_iterable(v) and "type" not in v):
                default_config[k] = config[k]

        # If the configuration describes an object, build it
        if "type" in config:
            # Retrieve the class from registry
            #assert "type" in config, f"The configuration must include an entry 'type' containing the type of the object to be built."
            type_name: str = config["type"]
            assert type_name in self.registry, f"The class `{type_name}` is not declared in the registry. To get the list of registered classes, use: 'print(registry.keys())'."
            cls: Any = self.registry[type_name]
            assert issubclass(cls, CreatableFromConfig), f"The class `{cls}` must inherit from `CreatableFromConfig` to be built."
            # Build object
            obj: Any = cls.from_config(config, **default_params, **kwargs)
            # If 'obj' possess a name, add it to the factory
            name: Optional[str] = self._get_name(obj, None, config)
            if name is not None and len(name) > 0 and not name in self:
                self[name] = obj
            return obj

        else: # Return all built objects
            return built_objs



    def build(self, config: Mapping[str, Any], default_params: Mapping[str, Any] = {}, **kwargs: Any) -> Any:
        """Create and return an object from `self.registry` based on
        configuration entries from `self` and `config`. The class of the created
        object must inherit from the Class `CreatableFromConfig`.
        The object is created by iteratively executing the `from_config` methods.
        The class of the object must be specified in configuration entry 'type'.

        If a configuration entry contains a sub-entry 'type', it also created by the factory.
        If it also contains a sub-entry 'name', it is added to `self` with key 'name', and accessible
        through `self[name]`.
        
        Parameters
        ----------
        :param config: Mapping[str, Any]
            The mapping containing configuration entries of the built object.
        :param default_params:
            Additional configuration entries to send to the `from_config` class method when creating the main object and all sub-objects.
        :param kwargs: Any
            Additional configuration entries to send to the `from_config` class method when creating the main object.
        """
        if "name" in config and config["name"] in self:
            return self[config["name"]]
        final_config: MutableMapping[str, Any] = dict(copy.deepcopy(config))
        return self._build_internal(final_config, default_params, **kwargs)



# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.


__all__ = ["Benchmark"]

from qdpy.base import *


class Benchmark(object):
    fn: Callable
    ind_domain: DomainLike
    fitness_domain: Sequence[DomainLike]
    features_domain: Sequence[DomainLike]
    default_task: str
    def __init__(self, fn: Callable, ind_domain: DomainLike, fitness_domain: Sequence[DomainLike], features_domain: Sequence[DomainLike], default_task: str = "minimisation"):
        self.fn = fn # type: ignore
        self.ind_domain = ind_domain
        self.fitness_domain = fitness_domain
        self.features_domain = features_domain
        self.default_task = default_task


# MODELINE "{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""Old preliminary code. It is not documented and will be removed in the future."""


########### IMPORTS ########### {{{1
import math
import numpy as np
import sys
import copy
import os
import pickle
import time
import gc
from scipy.spatial.distance import euclidean
from sklearn.cluster import KMeans
from sklearn.gaussian_process import GaussianProcessRegressor
import itertools
from timeit import default_timer as timer

#import matplotlib as mpl
#mpl.use('Agg')
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

from functools import reduce
import operator
import random
import warnings

import qdpy.hdsobol as hdsobol
from qdpy.utils import *


########### INITIALISATION AND MUTATIONS ########### {{{1
@jit(nopython=True)
def mutateUniformInt(ind, mutationPb, indBounds):
    for i in range(len(ind)):
        if np.random.random() < mutationPb:
            ind[i] = np.random.randint(indBounds[0], indBounds[1])

@jit(nopython=True)
def mutateUniform(ind, mutationPb, indBounds):
    for i in range(len(ind)):
        if np.random.random() < mutationPb:
            ind[i] = np.random.uniform(indBounds[0], indBounds[1])

@jit(nopython=True)
def generateUniformInt(dimension, indBounds, nb):
    res = []
    for i in range(nb):
        res.append(np.random.randint(indBounds[0], indBounds[1] + 1, dimension))
    return res

@jit(nopython=True)
def generateUniform(dimension, indBounds, nb):
    res = []
    for i in range(nb):
        res.append(np.random.uniform(indBounds[0], indBounds[1], dimension))
    return res

@jit(nopython=True)
def generateSparseUniform(dimension, indBounds, nb, sparsity):
    res = []
    for i in range(nb):
        base = np.random.uniform(indBounds[0], indBounds[1], dimension)
        mask = np.random.uniform(0., 1., dimension)
        base[mask < sparsity] = 0.
        res.append(base)
    return res

@jit(nopython=True)
def generateSobol(dimension, indBounds, nb):
    res = hdsobol.gen_sobol_vectors(nb+1, dimension)
    res = res * (indBounds[1] - indBounds[0]) + indBounds[0]
    return res

@jit(nopython=True)
def generateBinarySobol(dimension, indBounds, nb, cutoff = 0.50):
    res = hdsobol.gen_sobol_vectors(nb+1, dimension)
    res = (res > cutoff).astype(int)
    return res






########### MAP-Elites ########### {{{1
class MapElites(object):
    def __init__(self, dimension, evalFn, nbBins, featuresBounds = [(0., 1.)], initBatchSize = 120, batchSize=40, nbIterations = 10, indBounds = (0, 100), mutationPb = 0.2, savePeriod = 0, logBasePath = ".", reevalTimeout = None, mutate = None, initiate = None, iterationFilenames = "iteration-%i.p", finalFilename = "final.p", fitnessBounds = (0., 1.), parallelismType = "multiprocessing", completelyNewGenomePb = 0.0):
        self.dimension = dimension
        self.evalFn = evalFn
        self.nbBins = nbBins
        self.featuresBounds = featuresBounds
        self.initBatchSize = initBatchSize
        self.nbIterations = nbIterations
        self.batchSize = batchSize
        self.indBounds = indBounds
        self.mutationPb = mutationPb
        self.savePeriod = savePeriod
        self.logBasePath = logBasePath
        self.reevalTimeout = reevalTimeout
        self.mutate = mutate
        self.initiate = initiate
        self.iterationFilenames = iterationFilenames
        self.finalFilename = finalFilename
        self.fitnessBounds = fitnessBounds
        self.callEvalFnOnEntierBatch = False
        self.parallelismType = parallelismType
        self.completelyNewGenomePb = completelyNewGenomePb
        self.reinit()

    def __del__(self):
        self._closePool()

    def __getstate__(self):
        odict = self.__dict__.copy()
        if 'pool' in odict:
            del odict['pool']
        del odict['map']
        return odict

    def _closePool(self):
        if hasattr(self, 'pool'):
            self.pool.close()
            self.pool.terminate()

    def _defaultMutate(self, ind):
        #return mutateUniformInt(ind, self.mutationPb, self.indBounds)
        return mutateUniformInt

    def _initParallelism(self):
        if self.parallelismType == "none":
            #import itertools
            self.map = map #itertools.starmap
        elif self.parallelismType == "multiprocessing":
            import multiprocessing
            self._closePool()
            self.pool = multiprocessing.Pool()
            self.map = self.pool.map #self.pool.starmap
        elif self.parallelismType == "scoop":
            import scoop
            self.map = scoop.futures.map
        else:
            raise Exception("Unknown parallelismType: '%s'" % self.parallelismType)


    def reinit(self):
        if not hasattr(self, '_infoToSave'):
            self._infoToSave = {}
        self._initParallelism()
        self._unfinishedEvals = 0
        self._evalId = itertools.count()
        if not self.mutate:
            self.mutate = self._defaultMutate
        if not self.initiate:
            self.initiate = self._defaultInitiation
        self.totalElapsed = 0.

        self.initPop = []
        assert(self.nbIterations > 0)
        self.currentIteration = 0
        self.currentEvaluation = 0
        if self.nbBins == None:
            self.nbBins = np.repeat(10, len(self.featuresBounds))
        assert(len(self.featuresBounds) == len(self.nbBins))
        #self._binsSize = [(self.featuresBounds[1] - self.featuresBounds[0]) / float(b) for b in self.nbBins] 
        self._binsSize = [(self.featuresBounds[bi][1] - self.featuresBounds[bi][0]) / float(self.nbBins[bi]) for bi in range(len(self.nbBins))]
        assert(len(self.nbBins))
        self.performances = np.full(self.nbBins, self.fitnessBounds[0]) #np.zeros(shape=self.nbBins)
        self.features = np.zeros(shape=list(self.nbBins) + [len(self.nbBins)])
        self.solutions = {}
        self.bestInIteration = None
        self.bestInIterationFitness = self.fitnessBounds[0]
        self.bestEver = None
        self.bestEverFitness = self.fitnessBounds[0]
        self.containerSize = reduce(operator.mul, self.nbBins)
        self.nbEmptyBins = self.containerSize

    def _valsToElitesIndex(self, vals):
        index = []
        assert(len(vals) == len(self.nbBins))
        for i in range(len(vals)):
            normalisedVal = vals[i] - self.featuresBounds[i][0]
            if normalisedVal == self.featuresBounds[i][1] - self.featuresBounds[i][0]:
                partial = self.nbBins[i] - 1
            elif normalisedVal > self.featuresBounds[i][1] - self.featuresBounds[i][0]:
                print("WARNING: feature %i out of bounds: %s" % (i, vals))
                return None
            else:
                partial = int(normalisedVal / self._binsSize[i])
            index.append(partial)
        return tuple(index)

    def _asyncEvalFnWrapper(self, ind, expeId):
        def startEval(newInd):
            startTime = timer()
            res = self.evalFn(ind, expeId)
            endTime = timer()
            elapsed = endTime - startTime
            return elapsed, res
        totalElapsed = 0.
        currentInd = ind
        while(True): # XXX bad !
            elapsed, res = startEval(currentInd)
            totalElapsed += elapsed
            if self.reevalTimeout and self.reevalTimeout > totalElapsed:
                fitness = res[0]
                if fitness <= 0.000001:
                    # Perform a reevaluation
                    print("Performed a Reevaluation after:%f sec" % totalElapsed)
                    self.mutate(currentInd, self.mutationPb, self.indBounds)
                    sys.stdout.flush()
                    continue
                else:
                    return currentInd, res
            else:
                return currentInd, res

    def _evalFnWrapper(self, vals):
        ind, expeId = vals
        return self.evalFn(ind, expeId)

    def _evaluatePop(self, pop):
        startTime = timer()
        pop = np.unique(pop, axis=0)
        listEvalIds = list(next(self._evalId) for _ in pop)
        # Evaluate the individuals
        #popFitnesses = np.array(self.map(self.evalFn, pop)) #self.toolbox.map(self.toolbox.evaluate, pop)
        #popFitnesses = np.array(self.map(self.evalFn, list(zip(pop, listEvalIds)))) #self.toolbox.map(self.toolbox.evaluate, pop)
        #resFromEvals = self.map(self._evalFnWrapper, list(zip(pop, listEvalIds)))
        #popFinalInds, popFitnesses = zip(*resFromEvals)

        if self.callEvalFnOnEntierBatch:
            popFitnesses = np.array(list(self.evalFn(pop, listEvalIds)))
        else:
            #popFitnesses = self.map(self.evalFn, list(zip(pop, listEvalIds)))
            popFitnesses = self.map(self._evalFnWrapper, list(zip(pop, listEvalIds)))
        #nbMaxEvalsPerRound = 1000
        #if len(listEvalIds) > nbMaxEvalsPerRound:
        #    nbRounds = int(len(listEvalIds) / nbMaxEvalsPerRound)
        #    popFitnesses = []
        #    for i in range(nbRounds):
        #        fitList = self.map(self.evalFn, list(zip(pop[i * nbMaxEvalsPerRound: (i+1) * nbMaxEvalsPerRound], listEvalIds[i * nbMaxEvalsPerRound: (i+1) * nbMaxEvalsPerRound])))
        #        popFitnesses += fitList
        #        # Perform garbage collection
        #        while gc.collect() > 0:
        #            pass
        #else:
        #    popFitnesses = self.map(self.evalFn, list(zip(pop, listEvalIds)))
        while gc.collect() > 0:
            pass

        popFitnesses = np.array(list(popFitnesses))
        popPerformances = popFitnesses[:,0]
        popFeatures = popFitnesses[:,1:]
        # Record individuals
        for indIndex in range(len(pop)):
            ind = pop[indIndex]
            elitesIndex = self._valsToElitesIndex(popFeatures[indIndex])
            if elitesIndex == None:
                continue
            if self.performances[elitesIndex] < popPerformances[indIndex]:
                self.solutions[elitesIndex] = ind
                self.performances[elitesIndex] = popPerformances[indIndex]
                self.features[elitesIndex] = popFeatures[indIndex]
            if popPerformances[indIndex] > self.bestInIterationFitness:
                self.bestInIterationFitness = popPerformances[indIndex]
                self.bestInIteration = ind
        if self.bestInIterationFitness > self.bestEverFitness:
            self.bestEverFitness = self.bestInIterationFitness
            self.bestEver = self.bestInIteration
        self.nbEmptyBins = self.containerSize - len(self.solutions.keys())
        endTime = timer()
        elapsed = endTime - startTime
        #return popFinalInds, self.bestInIterationFitness, elapsed
        return self.bestInIterationFitness, elapsed


    def _evaluateIndAsync(self, ind = None):
        if ind is None:
            currentElites = list(self.solutions.values())
            if not len(currentElites):
                raise ValueError("No elites were found in initial batch !")
            ind = copy.deepcopy(currentElites[np.random.randint(0, len(currentElites))])
            self.mutate(ind, self.mutationPb, self.indBounds)
        #print("DEBUGasync: ", ind)
        evalId = next(self._evalId)
        self._unfinishedEvals += 1
        asyncResFromEval = self.pool.apply_async(self._asyncEvalFnWrapper, [ind, evalId], callback = self._evaluateIndAsyncCallback)
        return asyncResFromEval

    def _evaluateIndAsyncCallback(self, param):
        ind, res = param
        #print("DEBUGasyncC1: ", res)
        performance = res[0]
        features = res[1:]
        elitesIndex = self._valsToElitesIndex(features)
        if self.performances[elitesIndex] < performance:
            self.solutions[elitesIndex] = ind
            self.performances[elitesIndex] = performance
            self.features[elitesIndex] = features
        if performance > self.bestInIterationFitness:
                self.bestInIterationFitness = performance
                self.bestInIteration = ind
        # Update bestever
        if self.bestInIterationFitness > self.bestEverFitness:
            self.bestEverFitness = self.bestInIterationFitness
            self.bestEver = self.bestInIteration
        self.nbEmptyBins = self.containerSize - len(self.solutions.keys())

        self._unfinishedEvals -= 1
        self.currentEvaluation += 1
        #print("DEBUGasyncC2: ", self._unfinishedEvals, self.currentEvaluation, (self.currentEvaluation - self.initBatchSize) % self.batchSize)
        sys.stdout.flush()
        if self.currentEvaluation == self.initBatchSize:
            # Verify if the initial batch is finished
            print("#0 bestInBatch:%f bestEver:%f" % (self.bestInIterationFitness, self.bestInIterationFitness))
            sys.stdout.flush()
        elif self.currentEvaluation > self.initBatchSize and (self.currentEvaluation - self.initBatchSize) % self.batchSize == 0:
            # Verify if this iteration is finished
            self.currentIteration += 1
            print("#%i bestInBatch:%f bestEver:%f" % (self.currentIteration, self.bestInIterationFitness, self.bestEverFitness))
            sys.stdout.flush()
            if self.savePeriod and np.mod(self.currentIteration, self.savePeriod) == 0:
                #print("DEBUGasyncC3: ", self.iterationFilenames, self.currentIteration)
                self.save(os.path.join(self.logBasePath, self.iterationFilenames % self.currentIteration))

        # Verify if a new eval must be launched
        if self.currentEvaluation + self._unfinishedEvals >= self.initBatchSize + self.nbIterations * self.batchSize:
            if self._unfinishedEvals == 0:
                # All evaluation are finished
                self.save(os.path.join(self.logBasePath, self.finalFilename))
        else:
            # Launch new eval
            self._evaluateIndAsync()



    def _defaultInitiation(self, dimension, indBounds, nb):
        return generateUniformInt(dimension, self.indBounds, nb)

    def _generateInitBatch(self):
        self.initPop = np.unique(self.initiate(self.dimension, self.indBounds, self.initBatchSize), axis=0)
        #self.initPop = []
        #for i in range(self.initBatchSize):
        #    #self.initPop.append(np.random.randint(self.indBounds[0], self.indBounds[1], self.dimension))
        #    #self.initPop.append(np.random.uniform(self.indBounds[0], self.indBounds[1], self.dimension))
        #    self.initPop.append(self.initiate(self.dimension, self.indBounds))
        return self.initPop

    def _iterationMessage(self, prefixLogs, iteration, batch, bestInIterationFitness, bestEverFitness, elapsed):
        #print("%s%i batchSize:%i bestInBatch:%f bestEver:%f elapsed:%f" % (prefixLogs, iteration, len(batch), bestInIterationFitness, bestEverFitness, elapsed))
        print("%s%i batchSize:%i bestEver:%f emptyBins:%i/%i elapsed:%f" % (prefixLogs, iteration, len(batch), bestEverFitness, self.nbEmptyBins, self.containerSize, elapsed))

    def _genNewIndFromElites(self):
        if self.completelyNewGenomePb > 0.0 and np.random.random() < self.completelyNewGenomePb:
            return self.initiate(self.dimension, self.indBounds, 1)[0]
        else:
            currentElites = list(self.solutions.values())
            if not len(currentElites):
                raise ValueError("No elites were found !")
            ind = copy.deepcopy(currentElites[np.random.randint(0, len(currentElites))])
            self.mutate(ind, self.mutationPb, self.indBounds)
            return ind

    def run(self, initBatch = None, nbIterations = None, disableLogs = False, prefixLogs = "#"):
        startTime = timer()
        if initBatch:
            self.initPop = initBatch
        else:
            self._generateInitBatch()
        if nbIterations == None:
            nbIterations = self.nbIterations
        #self.initPop, self.bestInIterationFitness, elapsed = self._evaluatePop(self.initPop)
        self.bestInIterationFitness, elapsed = self._evaluatePop(self.initPop)
        self.bestEverFitness = self.bestInIterationFitness
        self.currentEvaluation = len(self.initPop)
        if not disableLogs:
            self._iterationMessage(prefixLogs, 0, self.initPop, self.bestInIterationFitness, self.bestEverFitness, elapsed)
        sys.stdout.flush()
        for iteration in range(1, nbIterations):
            self.currentIteration = iteration
            self.currentEvaluation += self.batchSize
            currentElites = list(self.solutions.values())
            if not len(currentElites):
                raise ValueError("No elites were found in initial batch !")
            newPop = []
            for i in range(self.batchSize):
                ind = self._genNewIndFromElites()
                newPop.append(ind)
            #newPop, self.bestInIterationFitness, elapsed = self._evaluatePop(newPop)
            self.bestInIterationFitness, elapsed = self._evaluatePop(newPop)
            #print("%s%i bestInBatch:%f bestEver:%f elapsed:%f elapsedPerInd:%f" % (prefixLogs, iteration, self.bestInIterationFitness, self.bestEverFitness, elapsed, elapsed / len(newPop)))
            if not disableLogs:
                self._iterationMessage(prefixLogs, iteration, newPop, self.bestInIterationFitness, self.bestEverFitness, elapsed)
            sys.stdout.flush()
            if not disableLogs and self.savePeriod and np.mod(self.currentIteration, self.savePeriod) == 0:
                self.save(os.path.join(self.logBasePath, self.iterationFilenames % iteration))
            # Perform garbage collection
            while gc.collect() > 0:
                pass
        if not disableLogs and self.finalFilename:
            self.save(os.path.join(self.logBasePath, self.finalFilename))
        endTime = timer()
        self.totalElapsed += endTime - startTime
        if not disableLogs and nbIterations > 1:
            print("Total elapsed:%f" % (self.totalElapsed))
        return (self.bestEver, self.bestEverFitness)


    def runAsync(self, initBatch = None):
        if initBatch:
            self.initPop = initBatch
        else:
            self._generateInitBatch()
        evalsFuturesList = []
        for i in range(len(self.initPop)):
            evalsFuturesList.append(self._evaluateIndAsync(self.initPop[i]))
            #self._evaluateIndAsync(self.initPop[i])
        # Main loop
        while(True):
            #print("DEBUG1:", self.currentEvaluation, self.initBatchSize + self.nbIterations * self.batchSize)
            #print("DEBUG1b: ", [x.ready() for x in evalsFuturesList])
            if self.currentEvaluation >= self.initBatchSize + self.nbIterations * self.batchSize:
                break
            time.sleep(1)
        return (self.bestEver, self.bestEverFitness)


    def generateOutputDict(self):
        outputDict = {}
        outputDict['performances'] = self.performances
        outputDict['features'] = self.features
        outputDict['solutions'] = self.solutions
        outputDict['dimension'] = self.dimension
        outputDict['nbBins'] = self.nbBins
        outputDict['featuresBounds'] = self.featuresBounds
        outputDict['initBatchSize'] = self.initBatchSize
        outputDict['nbIterations'] = self.nbIterations
        outputDict['batchSize'] = self.batchSize
        outputDict['indBounds'] = self.indBounds
        outputDict['mutationPb'] = self.mutationPb
        outputDict['currentIteration'] = self.currentIteration
        outputDict['initPop'] = self.initPop
        outputDict['bestEver'] = self.bestEver
        outputDict['bestEverFitness'] = self.bestEverFitness
        outputDict = {**outputDict, **self._infoToSave}
        return outputDict

    def save(self, outputFile):
        outputDict = self.generateOutputDict()
        with open(outputFile, "wb") as f:
            pickle.dump(outputDict, f)

    def addSavingInfo(self, key, value):
        self._infoToSave[key] = value



########### CVT-MAP-Elites ########### {{{1

#@jit(nopython=True)
@jit(nopython=True, parallel=True)
def _cvtMapElites_valsToElitesIndex(clusterCenters, vals):
    dists = np.empty((clusterCenters.shape[0]))
    for i in range(len(dists)):
        dists[i] = math.sqrt(np.sum(np.square(clusterCenters[i] - vals)))
    return (np.argmin(dists),)


class CVTMapElites(MapElites):
    def __init__(self, dimension, evalFn, nbBins, nbClusters, featuresBounds = [(0.0, 1.0)], initBatchSize = 120, batchSize=40, nbIterations = 10, indBounds = (0, 100), mutationPb = 0.2, savePeriod = 0, logBasePath = ".", nbSampledPoints = 50000, reevalTimeout = 5., mutate = None, initiate = None, iterationFilenames = "iteration-%i.p", finalFilename = "final.p", fitnessBounds = (0., 1.), parallelismType = "multiprocessing", completelyNewGenomePb = 0.0):
        self.nbClusters = nbClusters
        self.nbSampledPoints = nbSampledPoints
        super().__init__(dimension, evalFn, nbBins, featuresBounds, initBatchSize, batchSize, nbIterations, indBounds, mutationPb, savePeriod, logBasePath, reevalTimeout, mutate, initiate, iterationFilenames, finalFilename, fitnessBounds, parallelismType, completelyNewGenomePb=completelyNewGenomePb)

    def reinit(self):
        super().reinit()
        self.performances = np.zeros(shape=self.nbClusters)
        self.features = np.zeros(shape=[self.nbClusters, len(self.nbBins)])
        # Init clusters
        sample = np.random.uniform(self.featuresBounds[0][0], self.featuresBounds[0][1], (self.nbSampledPoints, len(self.nbBins)))
        kmeans = KMeans(init="k-means++", n_clusters=self.nbClusters, n_init=1, n_jobs=1, verbose=0)
        kmeans.fit(sample)
        self.clusterCenters = kmeans.cluster_centers_
        self.containerSize = self.nbClusters
        self.nbEmptyBins = self.containerSize

    def _valsToElitesIndex(self, vals):
        return _cvtMapElites_valsToElitesIndex(self.clusterCenters, vals)
        #closestDist = float("inf")
        #closestCluster = 0
        #for i in range(len(self.clusterCenters)):
        #    dist = euclidean(self.clusterCenters[i], vals)
        #    if dist < closestDist:
        #        closestDist = dist
        #        closestCluster = i
        #return (closestCluster,)

    def generateOutputDict(self):
        outputDict = super().generateOutputDict()
        outputDict['nbClusters'] = self.nbClusters
        outputDict['nbSampledPoints'] = self.nbSampledPoints
        outputDict['clusterCenters'] = self.clusterCenters
        outputDict = {**outputDict, **self._infoToSave}
        return outputDict



########### SAIL ########### {{{1
class SAIL(object):
    def __init__(self, dimension, evalFn, nbBins, illuminationAlgo, acquisitionMapAlgo, predictionMapAlgo, featuresBounds = [(0.0, 1.0)], initBatchSize = 120, batchSize=40, indBounds = (0, 100), mutationPb = 0.2, savePeriod = 0, logBasePath = ".", mutate = None, initiate = None, nbAcquisitionRounds = 10, ucbStdDevFactor = 1.0, nbPreciseEvalsPerAcquisitionRound = 10, instanceName = "", fitnessBounds = (0., 1.), finalFilename = "final.p", parallelismType = "multiprocessing", maintainBatchSizeWithMutatedInd = False, preciseEvalsSelectionType = "sobol"):
        self.dimension = dimension
        self.evalFn = evalFn
        self.nbBins = nbBins
        self.illuminationAlgo = illuminationAlgo
        self.acquisitionMapAlgo = acquisitionMapAlgo
        self.predictionMapAlgo = predictionMapAlgo
        self.featuresBounds = featuresBounds
        self.initBatchSize = initBatchSize
        self.batchSize = batchSize
        self.indBounds = indBounds
        self.mutationPb = mutationPb
        self.savePeriod = savePeriod
        self.logBasePath = logBasePath
        self.mutate = mutate
        self.initiate = initiate
        self.nbAcquisitionRounds = nbAcquisitionRounds
        self.ucbStdDevFactor = ucbStdDevFactor
        self.nbPreciseEvalsPerAcquisitionRound = nbPreciseEvalsPerAcquisitionRound
        self.instanceName = instanceName
        self.fitnessBounds = fitnessBounds
        self.finalFilename = finalFilename
        self.parallelismType = parallelismType
        self.maintainBatchSizeWithMutatedInd = maintainBatchSizeWithMutatedInd
        self.preciseEvalsSelectionType = preciseEvalsSelectionType
        self.reinit()

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['gp']
        return odict


    def reinit(self):
        self._infoToSave = {}
        self.acquisitionMapAlgo.evalFn = self._acquisitionEvalFn
        self.predictionMapAlgo.evalFn = self._predictionEvalFn
        self.acquisitionMapAlgo.parallelismType = "none" # Disable multiprocessing
        self.predictionMapAlgo.parallelismType = "none" # Disable multiprocessing
        self.currentRound = 0


    def _acquisitionEvalFn(self, inds, expeIds):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            y_mean, y_std = self.gp.predict(np.array(inds), return_std = True)
        ucbVal = y_mean + self.ucbStdDevFactor * y_std.reshape((-1, 1))
        # Normalise ucbVal
        ucbVal[:,0][ucbVal[:,0] < self.fitnessBounds[0]] = self.fitnessBounds[0]
        ucbVal[:,0][ucbVal[:,0] > self.fitnessBounds[1]] = self.fitnessBounds[1]
        for i in range(len(self.featuresBounds)):
            ucbVal[:,i+1][ucbVal[:,i+1] < self.featuresBounds[i][0]] = self.featuresBounds[i][0]
            ucbVal[:,i+1][ucbVal[:,i+1] > self.featuresBounds[i][1]] = self.featuresBounds[i][1]
        return copy.deepcopy(ucbVal)

    def _predictionEvalFn(self, inds, expeIds):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            y_mean, y_std = self.gp.predict(np.array(inds), return_std = True)
        # Normalise y_mean
        y_mean[:,0][y_mean[:,0] < self.fitnessBounds[0]] = self.fitnessBounds[0]
        y_mean[:,0][y_mean[:,0] > self.fitnessBounds[1]] = self.fitnessBounds[1]
        for i in range(len(self.featuresBounds)):
            y_mean[:,i+1][y_mean[:,i+1] < self.featuresBounds[i][0]] = self.featuresBounds[i][0]
            y_mean[:,i+1][y_mean[:,i+1] > self.featuresBounds[i][1]] = self.featuresBounds[i][1]
        return copy.deepcopy(y_mean)

    def _createGaussianRegressor(self):
        # Create dataset
        elitesIndexes = list(self.illuminationAlgo.solutions.keys())
        assert(len(elitesIndexes))
        dataX = []
        dataY = []
        for index in elitesIndexes:
            dataX.append(self.illuminationAlgo.solutions[index])
            perfAndFeatures = [self.illuminationAlgo.performances[index]] + list(self.illuminationAlgo.features[index])
            dataY.append(perfAndFeatures)
        dataX = np.array(dataX)
        dataY = np.array(dataY)

        # Fit gaussian process on the dataset
        #print("DEBUGii", dataX, dataY, self.fitnessBounds, np.min(self.illuminationAlgo.performances), elitesIndexes)
        self.gp = GaussianProcessRegressor()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.gp.fit(dataX, dataY)
        return self.gp

    def _genNewIndFromElites(self):
        return self.illuminationAlgo._genNewIndFromElites()

    def _produceAcquisitionMap(self, disableLogs = True, prefixLogs = "#"):
        for roundId in range(1, self.nbAcquisitionRounds):
            startTimeAcquisition = timer()
            self.currentRound += 1
            # Reset acquisitionMapAlgo
            self.acquisitionMapAlgo.reinit()

            self._createGaussianRegressor()
            pop = list(self.illuminationAlgo.solutions.values())
            #print("DEBUGAC:", len(pop))
            # Launch map-elites using the gp as surrogate model
            if self.savePeriod and np.mod(roundId, self.savePeriod) == 0:
                self.acquisitionMapAlgo.iterationFilenames = "iterationAcquisition-%i-%s_%s.p" % (roundId, "%i", self.instanceName)
                self.acquisitionMapAlgo.finalFilename = "finalAcquisition-%i_%s.p" % (roundId, self.instanceName)
            else:
                self.acquisitionMapAlgo.iterationFilenames = None
                self.acquisitionMapAlgo.finalFilename = None
            self.acquisitionMapAlgo.initBatchSize = len(pop)
            bestEver, bestEverFitness = self.acquisitionMapAlgo.run(pop, disableLogs = disableLogs, prefixLogs = "Acquisition%i-%s" % (roundId, prefixLogs))

            # Select bins from acquisitionMapAlgo map
            #acquisitionSolutions = list(self.acquisitionMapAlgo.solutions.values())
            acquisitionSolutions = set([tuple(x) for x in self.acquisitionMapAlgo.solutions.values()])
            acquisitionSolutions -= set([tuple(x) for x in pop])
            if len(acquisitionSolutions):
                acquisitionSolutions = np.unique([np.array(x) for x in list(acquisitionSolutions)], axis=0)
                if len(acquisitionSolutions) <= self.nbPreciseEvalsPerAcquisitionRound:
                    selectedSolutions = acquisitionSolutions
                else:
                    # Select how the index is generated (random uniform sampling vs sobol sampling)
                    if self.preciseEvalsSelectionType == "sobol":
                        indexList = np.unique((hdsobol.gen_sobol_vectors(self.nbPreciseEvalsPerAcquisitionRound+1, 1) * len(acquisitionSolutions)).astype(int))
                    elif self.preciseEvalsSelectionType == "uniform":
                        indexList = random.sample(range(len(acquisitionSolutions)), self.nbPreciseEvalsPerAcquisitionRound)
                    else:
                        raise ValueError("Unknown value '%s' for entry 'preciseEvalsSelectionType'" % self.preciseEvalsSelectionType)
                    selectedSolutions = [acquisitionSolutions[x] for x in indexList]

                pop = list(selectedSolutions)
                if self.maintainBatchSizeWithMutatedInd:
                    for i in range(self.batchSize - len(selectedSolutions)):
                        pop.append(self._genNewIndFromElites())

                bestInIterationFitness, elapsedIllumination = self.illuminationAlgo._evaluatePop(pop)
                endTimeAcquisition = timer()
                elapsed = elapsedIllumination + endTimeAcquisition - startTimeAcquisition
                self.illuminationAlgo._iterationMessage(prefixLogs, roundId, pop, self.illuminationAlgo.bestInIterationFitness, self.illuminationAlgo.bestEverFitness, elapsed)
                if self.savePeriod and np.mod(roundId, self.savePeriod) == 0:
                    self.illuminationAlgo.iterationFilenames = "iteration-%i-%s_%s.p" % (roundId, "%i", self.instanceName)
                    self.illuminationAlgo.save(os.path.join(self.logBasePath, self.illuminationAlgo.iterationFilenames % roundId))

            self.bestEverFitness = self.illuminationAlgo.bestEverFitness
            self.bestEver = self.illuminationAlgo.bestEver
            self.containerSize = self.illuminationAlgo.containerSize
            self.nbEmptyBins = self.illuminationAlgo.nbEmptyBins
            self.performances = self.illuminationAlgo.performances
            self.solutions = self.illuminationAlgo.solutions
            self.features = self.illuminationAlgo.features

            # Perform garbage collection
            while gc.collect() > 0:
                pass



    def _producePredictionMap(self, disableLogs = True, prefixLogs = "#"):
        self.predictionMapAlgo.reinit()
        self._createGaussianRegressor()
        pop = list(self.illuminationAlgo.solutions.values())
        # Launch map-elites using the gp as surrogate model
        self.predictionMapAlgo.iterationFilenames = "iterationPrediction-%s_%s.p" % ("%i", self.instanceName)
        self.predictionMapAlgo.finalFilename = "finalPrediction_%s.p" % self.instanceName
        self.predictionMapAlgo.initBatchSize = len(pop)
        bestEver, bestEverFitness = self.predictionMapAlgo.run(pop, disableLogs = disableLogs, prefixLogs = "Prediction-" + prefixLogs)

    def _updateConfig(self):
        def updateEntry(entry):
            self.illuminationAlgo.__dict__[entry] = self.__dict__[entry]
            self.acquisitionMapAlgo.__dict__[entry] = self.__dict__[entry]
            self.predictionMapAlgo.__dict__[entry] = self.__dict__[entry]
        list(map(updateEntry, ["initiate", "mutate", "logBasePath", "nbBins", "_infoToSave", "dimension", "featuresBounds", "indBounds", "fitnessBounds"]))
        self.illuminationAlgo.evalFn = self.evalFn
        self.illuminationAlgo.iterationFilenames = None
        self.illuminationAlgo.finalFilename = None
        self.illuminationAlgo.initBatchSize = self.initBatchSize
        self.acquisitionMapAlgo.callEvalFnOnEntierBatch = True
        self.predictionMapAlgo.callEvalFnOnEntierBatch = True
        self.illuminationAlgo.reinit()

    def run(self, initBatch = None, nbIterations = None, disableLogs = False, prefixLogs = "#"):
        self._updateConfig()
        # Generate initial population
        self.illuminationAlgo.run(initBatch = initBatch, nbIterations = 0, disableLogs = disableLogs, prefixLogs = prefixLogs) #, prefixLogs="Init-#")
        # Create maps
        if len(list(self.illuminationAlgo.solutions.keys())):
            self._produceAcquisitionMap(disableLogs = True, prefixLogs = prefixLogs)
            self._producePredictionMap(disableLogs = True, prefixLogs = prefixLogs)
        # Save
        #self.illuminationAlgo.finalFilename = "final_%s.p" % (self.instanceName)
        #self.illuminationAlgo.save(os.path.join(self.logBasePath, self.illuminationAlgo.finalFilename))
        #self.finalFilename = "final_%s.p" % (self.instanceName)
        if self.finalFilename:
            self.save(os.path.join(self.logBasePath, self.finalFilename))
        return (self.illuminationAlgo.bestEver, self.illuminationAlgo.bestEverFitness)


    def generateOutputDict(self):
        outputDict = {}
        outputDict['performances'] = self.performances
        outputDict['features'] = self.features
        outputDict['solutions'] = self.solutions
        outputDict['dimension'] = self.dimension
        outputDict['nbBins'] = self.nbBins
        outputDict['featuresBounds'] = self.featuresBounds
        outputDict['initBatchSize'] = self.initBatchSize
        outputDict['batchSize'] = self.batchSize
        outputDict['indBounds'] = self.indBounds
        outputDict['mutationPb'] = self.mutationPb
        outputDict['currentRound'] = self.currentRound
        outputDict['fitnessBounds'] = self.fitnessBounds
        outputDict['illuminationAlgo'] = self.illuminationAlgo.generateOutputDict()
        outputDict['acquisitionMapAlgo'] = self.acquisitionMapAlgo.generateOutputDict()
        outputDict['predictionMapAlgo'] = self.predictionMapAlgo.generateOutputDict()
        outputDict['bestEver'] = self.bestEver
        outputDict['bestEverFitness'] = self.bestEverFitness
        outputDict = {**outputDict, **self._infoToSave}
        return outputDict

    def save(self, outputFile):
        outputDict = self.generateOutputDict()
        with open(outputFile, "wb") as f:
            pickle.dump(outputDict, f)

    def addSavingInfo(self, key, value):
        self._infoToSave[key] = value





########### Factory ########### {{{1

class AlgorithmFactory(object):
    def __init__(self, algoType = "map-elites", nbDim = 50, evalFn = None, nbBins = None, nbClusters = 30, nbSampledPoints = 50000, featuresBounds = [(0., 1.)], indBounds = (0, 100), initBatchSize = 10000, batchSize = 4000, nbIterations = 500, mutationPb = 0.2, nbCenters = 20, savePeriod = 0, logBasePath = ".", reevalTimeout = None, mutate = None, initiate = None, iterationFilenames = "iteration-%i.p", finalFilename = "final.p",  instanceName = "", fitnessBounds = (0., 1.), parallelismType = "multiprocessing", maintainBatchSizeWithMutatedInd = False, completelyNewGenomePb = 0.0, preciseEvalsSelectionType = "sobol"):
        self.algoType = algoType
        self.nbDim = nbDim
        self.evalFn = evalFn
        self.nbBins = nbBins
        self.nbClusters = nbClusters
        self.nbSampledPoints = nbSampledPoints
        self.featuresBounds = featuresBounds
        self.indBounds = indBounds
        self.initBatchSize = initBatchSize
        self.batchSize = batchSize
        self.nbIterations = nbIterations
        self.mutationPb = mutationPb
        self.nbCenters = nbCenters
        self.savePeriod = savePeriod
        self.logBasePath = logBasePath
        self.reevalTimeout = reevalTimeout
        self.mutate = mutate
        self.initiate = initiate
        self.iterationFilenames = iterationFilenames
        self.finalFilename = finalFilename
        self.instanceName = instanceName
        self.fitnessBounds = fitnessBounds
        self.parallelismType = parallelismType
        self.maintainBatchSizeWithMutatedInd = maintainBatchSizeWithMutatedInd
        self.completelyNewGenomePb = completelyNewGenomePb
        self.preciseEvalsSelectionType = preciseEvalsSelectionType
        self.algo = None

    def update(self, **kwargs):
        for k,v in kwargs.items():
            if k in self.__dict__:
                self.__dict__[k] = v
            else:
                raise ValueError("Unknown entry: '%s'" % k)

    def build(self):
        if self.algoType == "map-elites":
            self.algo = MapElites(self.nbDim, self.evalFn, self.nbBins, self.featuresBounds, self.initBatchSize, self.batchSize, self.nbIterations, self.indBounds, self.mutationPb, self.savePeriod, self.logBasePath, self.reevalTimeout, self.mutate, self.initiate, self.iterationFilenames, self.finalFilename, fitnessBounds = self.fitnessBounds, parallelismType = self.parallelismType, completelyNewGenomePb = self.completelyNewGenomePb)
        elif self.algoType == "cvt-map-elites":
            self.algo = CVTMapElites(self.nbDim, self.evalFn, self.nbBins, self.nbClusters, self.featuresBounds, self.initBatchSize, self.batchSize, self.nbIterations, self.indBounds, self.mutationPb, self.savePeriod, self.logBasePath, self.nbSampledPoints, self.reevalTimeout, self.mutate, self.initiate, self.iterationFilenames, self.finalFilename, fitnessBounds = self.fitnessBounds, parallelismType = self.parallelismType, completelyNewGenomePb = self.completelyNewGenomePb)
        elif self.algoType == "sail":
            self.algo = SAIL(self.nbDim, self.evalFn, self.nbBins, self.illuminationAlgo, self.acquisitionMapAlgo, self.predictionMapAlgo, self.featuresBounds, self.initBatchSize, self.batchSize, self.indBounds, self.mutationPb, self.savePeriod, self.logBasePath, self.mutate, self.initiate, self.nbAcquisitionRounds, self.ucbStdDevFactor, self.nbPreciseEvalsPerAcquisitionRound, instanceName=self.instanceName, fitnessBounds = self.fitnessBounds, finalFilename = self.finalFilename, parallelismType = self.parallelismType, maintainBatchSizeWithMutatedInd = self.maintainBatchSizeWithMutatedInd, preciseEvalsSelectionType = self.preciseEvalsSelectionType)
        else:
            raise ValueError("Unknown algoType: '%s'" % self.algoType)
        return self.algo


    def fromConfig(self, config):
        def setIfExists(key):
            o = config.get(key)
            if o:
                self.__dict__[key] = o
        def buildIfExists(key):
            o = config.get(key)
            if o:
                fact = AlgorithmFactory()
                fact.fromConfig(o)
                fact.update(evalFn = self.evalFn, featuresBounds = self.featuresBounds, \
                        fitnessBounds = self.fitnessBounds, indBounds = self.indBounds, \
                        nbDim = self.nbDim, nbClusters = self.nbClusters, mutationPb = self.mutationPb, \
                        parallelismType = self.parallelismType, completelyNewGenomePb = self.completelyNewGenomePb)
                self.__dict__[key] = fact.build()
        setIfExists('parallelismType')
        setIfExists('algoType')
        setIfExists('nbDim')
        setIfExists('nbBins')
        setIfExists('nbClusters')
        setIfExists('nbSampledPoints')
        setIfExists('featuresBounds')
        setIfExists('indBounds')
        setIfExists('fitnessBounds')
        setIfExists('initBatchSize')
        setIfExists('batchSize')
        setIfExists('nbIterations')
        setIfExists('mutationPb')
        setIfExists('nbCenters')
        setIfExists('savePeriod')
        setIfExists('reevalTimeout')
        setIfExists('completelyNewGenomePb')
        # SAIL config
        setIfExists('nbAcquisitionRounds')
        setIfExists('ucbStdDevFactor')
        setIfExists('nbPreciseEvalsPerAcquisitionRound')
        setIfExists('maintainBatchSizeWithMutatedInd')
        setIfExists('preciseEvalsSelectionType')
        buildIfExists('illuminationAlgo')
        buildIfExists('acquisitionMapAlgo')
        buildIfExists('predictionMapAlgo')






########### Plots ########### {{{1

def plotMAP(performances, outputFilename, cmap, featuresBounds=[(0., 1.), (0., 1.)], fitnessBounds=(0., 1.), drawCbar = True, xlabel = "", ylabel = "", cBarLabel = "", nbTicks = 10):
    data = performances
    if len(data.shape) == 1:
        data = data.reshape((1, data.shape[0]))
    nbBins = data.shape

    figsize = [2.1 + 10. * nbBins[0] / (nbBins[0] + nbBins[1]), 1. + 10. * nbBins[1] / (nbBins[0] + nbBins[1])]
    aspect = "equal"
    if figsize[1] < 2:
        figsize[1] = 2.
        aspect = "auto"

    fig, ax  = plt.subplots(figsize=figsize)
    cax = ax.imshow(data.T, interpolation="none", cmap=cmap, vmin=fitnessBounds[0], vmax=fitnessBounds[1], aspect=aspect)
    #ax.set_aspect('equal')
    ax.invert_yaxis()

    if nbBins[0] > nbBins[1]:
        nbTicksX = nbTicks
        nbTicksY = int(nbTicksX * nbBins[1] / nbBins[0])
    elif nbBins[1] > nbBins[0]:
        nbTicksY = nbTicks
        nbTicksX = int(nbTicksY * nbBins[0] / nbBins[1])
    else:
        nbTicksX = nbTicksY = nbTicks
    if nbTicksX > nbBins[0] or nbTicksX < 1:
        nbTicksX = nbBins[0]
    if nbTicksY > nbBins[1] or nbTicksY < 1:
        nbTicksY = nbBins[1]

    # Set ticks
    ax.xaxis.set_tick_params(which='major', left=False, bottom=False, top=False, right=False)
    ax.yaxis.set_tick_params(which='major', left=False, bottom=False, top=False, right=False)
    xticks = list(np.arange(0, data.shape[0] + 1, data.shape[0] / nbTicksX))
    yticks = list(np.arange(0, data.shape[1] + 1, data.shape[1] / nbTicksY))
    plt.xticks(xticks, rotation='vertical')
    plt.yticks(yticks)
    deltaFeature0 = featuresBounds[0][1] - featuresBounds[0][0]
    deltaFeature1 = featuresBounds[1][1] - featuresBounds[1][0]
    ax.set_xticklabels([round(float(x / float(data.shape[0]) * deltaFeature0 + featuresBounds[0][0]), 2) for x in xticks], fontsize=20)
    ax.set_yticklabels([round(float(y / float(data.shape[1]) * deltaFeature1 + featuresBounds[1][0]), 2) for y in yticks], fontsize=20)

    # Draw grid
    ax.xaxis.set_tick_params(which='minor', direction="in", left=True, bottom=True, top=True, right=True)
    ax.yaxis.set_tick_params(which='minor', direction="in", left=True, bottom=True, top=True, right=True)
    ax.set_xticks(np.arange(-.5, data.shape[0], 1), minor=True)
    ax.set_yticks(np.arange(-.5, data.shape[1], 1), minor=True)
    ax.grid(which='minor', color=(0.8,0.8,0.8,0.5), linestyle='-', linewidth=0.1)

    ax.set_xlabel(xlabel, fontsize=25)
    ax.set_ylabel(ylabel, fontsize=25)

    if drawCbar:
        divider = make_axes_locatable(ax)
        #cax2 = divider.append_axes("right", size="5%", pad=0.15)
        cax2 = divider.append_axes("right", size=0.5, pad=0.15)
        cbar = fig.colorbar(cax, cax=cax2, format="%.2f")
        cbar.ax.tick_params(labelsize=20)
        cbar.ax.set_ylabel(cBarLabel, fontsize=22)

    ax.autoscale_view()
    plt.tight_layout()
    fig.savefig(outputFilename)





# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""TODO""" # TODO

__all__ = ["AlgorithmLogger", "TQDMAlgorithmLogger", "default_algorithm_logger"]


import os
from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, Mapping, MutableMapping, overload
import warnings
import numpy as np
import pandas as pd
import threading


from .base import *
from qdpy.utils import *
from qdpy.base import *
from qdpy.containers import *
from qdpy import tools



class AlgorithmLogger(Saveable):
    """TODO""" # TODO
    algorithms: MutableSequence[QDAlgorithmLike]
    iteration_filenames: Union[str, Callable]
    final_filename: str
    final_hdf_filename: Optional[str]
    save_period: int
    verbose: bool
    log_base_path: str
    config: Any
    saved_dict: Mapping

    _cols_names: Sequence[str] = ["alg_name", "iteration", "cont_size", "evals", "nb_updated", "avg", "std", "min", "max", "elapsed"]
    _cols_size: Sequence[int]
    _min_cols_size: int = 15
    _first_iteration: bool
    _current_iteration: int
    _current_evaluation: int

    #_cols_names_evals: Sequence[str] = ["alg_name", "iteration", "cont_size", "max", "elapsed"]
    _evals_data: MutableMapping[str, MutableSequence[Any]]
    _max_nb_objectives: int
    _iterations_data: MutableMapping[str, MutableSequence[Any]]
    _lock_evals_data: threading.Lock
    _lock_iterations_data: threading.Lock

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['_lock_evals_data']
        del odict['_lock_iterations_data']
        return odict


    def __init__(self, algorithms: Union[QDAlgorithmLike, Sequence[QDAlgorithmLike]] = [],
            iteration_filenames: Union[str, Callable] = "iteration-%i.p", final_filename: str = "final.p",
            final_hdf_filename: Optional[str] = None,
            save_period: int = 0, verbose: bool = True,
            log_base_path: str = ".", config: Any = None,
            saved_dict: Mapping = {}):
        self.iteration_filenames = iteration_filenames
        self.final_filename = final_filename
        self.final_hdf_filename = final_hdf_filename
        self.save_period = save_period
        self.verbose = verbose
        self.algorithms = []
        self.log_base_path = log_base_path
        self.config = config
        self.saved_dict = saved_dict
        self._cols_size = [len(x) for x in self._cols_names]
        self._first_iteration = False
        self._current_iteration = 0
        self._current_evaluation = 0
        # Initialise dataframes
        #self.evals = pd.DataFrame(index=np.arange(0, tot_budget), columns=self._cols_names_evals, dtype=object)
        #self.evals = pd.DataFrame(index=np.arange(0, tot_budget), columns=self._cols_names_evals)
        #self.evals = pd.DataFrame(columns=self._cols_names_evals)
        #self.iterations = pd.DataFrame(columns=self._cols_names)
        self._iterations_data = {k:[] for k in self._cols_names}
        self._lock_iterations_data = threading.Lock()
        #self._evals_data = {k:[] for k in self._cols_names_evals}
        self._evals_data = {}
        self._max_nb_objectives = 0
        self._lock_evals_data = threading.Lock()
        # Set algorithms to be monitored
        self.monitor(algorithms)


    @property
    def evals(self) -> pd.DataFrame:
        self._lock_evals_data.acquire()
        res = pd.DataFrame(self._evals_data)
        self._lock_evals_data.release()
        return res

    @property
    def iterations(self) -> pd.DataFrame:
        self._lock_iterations_data.acquire()
        res = pd.DataFrame(self._iterations_data)
        self._lock_iterations_data.release()
        return res


    def __get_saved_state__(self) -> Mapping[str, Any]:
        """Return a dictionary containing the relevant information to save.
        Only include information from `self.algorithms`, and evaluations and iterations DataFrames."""
        if len(self.algorithms) == 1 and isinstance(self.algorithms[0], Saveable):
            algos: Mapping[str, Any] = self.algorithms[0].__get_saved_state__()
        else:
            algos = {a.name: a for a in self.algorithms}
        return {**algos, "evals": self.evals, "iterations": self.iterations, "config": self.config, **self.saved_dict}


    def _output(self, *args):
        print(*args)

    def _gen_iteration_filename(self, iteration: Optional[int] = None) -> str:
        """Generate the filename of an iteration log file."""
        if isinstance(self.iteration_filenames, str):
            if iteration is None:
                raise ValueError("`iteration` must be provided.")
            return self.iteration_filenames % iteration
        else:
            return self.iteration_filenames()

    def _stats(self, algo: QDAlgorithmLike, batch_elapsed: float) -> Sequence[Any]:
        alg_name = algo.name
        #iteration = algo.current_iter - 1
        iteration = self._current_iteration
        cont_size = algo.container.size_str()
        evals = algo.nb_evaluations_in_iteration
        nb_updated = algo.nb_updated_in_iteration
        fitness_values = np.array([ind.fitness.values for ind in algo.container])
        avg_values = np.mean(fitness_values, axis=0) if len(fitness_values) else np.nan
        std_values = np.std(fitness_values, axis=0) if len(fitness_values) else np.nan
        min_values = np.min(fitness_values, axis=0) if len(fitness_values) else np.nan
        max_values = np.max(fitness_values, axis=0) if len(fitness_values) else np.nan
        res = [alg_name, iteration, cont_size, evals, nb_updated, avg_values, std_values, min_values, max_values, batch_elapsed]
        return res

    def _init_cols(self, content: Sequence[Any]) -> None:
        assert len(content) == len(self._cols_size)
        c_size: Sequence[int] = [len(f"{c}") for c in content]
        self._cols_size = [max(a, b, self._min_cols_size) for a, b in zip(self._cols_size, c_size)]


    def _vals_to_cols(self, content: Sequence[Any]) -> str:
        assert len(content) == len(self._cols_size)
        str_c: MutableSequence[str] = []
        for i, c in enumerate(content):
            str_c.append(("{0:<" + str(self._cols_size[i]) + "}").format(f"{c}"))
        return "\t".join(str_c)

    def _vals_to_cols_title(self, content: Sequence[Any]) -> str:
        return self._vals_to_cols(content)

    def _str_to_col(self, s: str, length: int) -> str:
        return ("{0:<" + str(length) + "}").format(s)


    def _tell(self, algo: QDAlgorithmLike, ind: IndividualLike) -> None:
        self._lock_evals_data.acquire()
        self._evals_data.setdefault('alg_name', []).append(algo.name)
        self._evals_data.setdefault('iteration', []).append(self._current_iteration)
        self._evals_data.setdefault('cont_size', []).append(algo.container.size)
        best_fitness: Any = algo.container.best_fitness
        best_fitness = list(best_fitness) if best_fitness is not None else [np.nan]
        for i in range(len(best_fitness)):
            self._evals_data.setdefault(f'max{i}', []).append(best_fitness[i])
        self._max_nb_objectives = max(self._max_nb_objectives, len(best_fitness))
        if self._max_nb_objectives > len(best_fitness):
            for i in range(len(best_fitness), self._max_nb_objectives - len(best_fitness)):
                self._evals_data.setdefault(f'max{i}', []).append(np.nan)
        self._evals_data.setdefault('elapsed', []).append(ind.elapsed)
        self._lock_evals_data.release()
        self._current_evaluation += 1

    def _iteration(self, algo: QDAlgorithmLike, batch_elapsed: float) -> None:
        if self.verbose:
            stats: Sequence[Any] = self._stats(algo, batch_elapsed)
            if self._first_iteration:
                self._first_iteration = False
                self._init_cols(stats)
                self._output(self._vals_to_cols_title(self._cols_names))
            self._output(self._vals_to_cols(stats))
        #finished_iter_nb: int = algo.current_iter - 1
        finished_iter_nb: int = self._current_iteration
        if self.save_period != 0 and finished_iter_nb % self.save_period == 0:
            if not (isinstance(self.iteration_filenames, str) and len(self.iteration_filenames) == 0):
                self.save(os.path.join(self.log_base_path, self._gen_iteration_filename(finished_iter_nb)))
        # Update iteration infos
        self._lock_iterations_data.acquire()
        for i in range(len(self._cols_names)):
            self._iterations_data[self._cols_names[i]].append(stats[i])
        self._lock_iterations_data.release()
        # Update current iteration nb
        self._current_iteration += 1

    def _started_optimisation(self, algo: QDAlgorithmLike) -> None:
        self._first_iteration = True

    def _finished_optimisation(self, algo: QDAlgorithmLike, optimisation_elapsed: float) -> None:
        if self.verbose:
            self._output(f"Finished optimisation using algorithm '{algo.name}'. Total elapsed: {optimisation_elapsed}.")
        if self.final_filename:
            self.save(os.path.join(self.log_base_path, self.final_filename))
        if self.final_hdf_filename:
            self.iterations.to_hdf(os.path.join(self.log_base_path, self.final_hdf_filename), key="iterations", mode="w")
            self.evals.to_hdf(os.path.join(self.log_base_path, self.final_hdf_filename), key="evals")

    def monitor(self, algorithms: Union[QDAlgorithmLike, Sequence[QDAlgorithmLike]]) -> None:
        """Monitor an algorithm or a Sequence of algorithms.
        It will the events of the optimisation process, namely when an iteration is finished,
        or when the `ask` or `tell` methods are called.
        
        Parameters
        ----------
        :param algorithms: Union[QDAlgorithmLike, Sequence[QDAlgorithmLike]]
            Either the algorithm to monitor, or a Sequence of algorithms to monitor.
        """
        if isinstance(algorithms, QDAlgorithmLike):
            algorithms = [algorithms]
        for a in algorithms:
            a.add_callback("tell", self._tell)
            a.add_callback("iteration", self._iteration)
            a.add_callback("started_optimisation", self._started_optimisation)
            a.add_callback("finished_optimisation", self._finished_optimisation)
            self.algorithms.append(a)

default_algorithm_logger = AlgorithmLogger(iteration_filenames = "", final_filename = "", verbose = True)


# Colorama is a python package to easily produce colored terminal outputs
try:
    import colorama
    _IMPORTED_COLORAMA = True
    colorama.init()
except Exception:
    _IMPORTED_COLORAMA = False


try:
    import tqdm
    _IMPORTED_TQDM = True
    class TQDMAlgorithmLogger(AlgorithmLogger):
        """TODO"""
        _tqdm_pbar: Optional[Any]

        def __getstate__(self):
            entries = {k:v for k,v in inspect.getmembers(self) if not k.startswith('_') and not inspect.ismethod(v)}
            return entries

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._tqdm_pbar = None

        def _output(self, *args):
            if self._tqdm_pbar is not None:
                self._tqdm_pbar.write("".join(args))

        def _started_optimisation(self, algo: QDAlgorithmLike) -> None:
            super()._started_optimisation(algo)
            if self._tqdm_pbar is not None:
                self._tqdm_pbar.close()
            self._tqdm_pbar = tqdm.tqdm(total=algo.budget, unit="eval", leave=False)

        def _finished_optimisation(self, algo: QDAlgorithmLike, optimisation_elapsed: float) -> None:
            super()._finished_optimisation(algo, optimisation_elapsed)
            if self._tqdm_pbar is not None:
                self._tqdm_pbar.close()
            self._tqdm_pbar = None

        def _tell(self, algo: QDAlgorithmLike, ind: IndividualLike) -> None:
            super()._tell(algo, ind)
            if self._tqdm_pbar is not None:
                self._tqdm_pbar.set_postfix(iteration=self._current_iteration)
                self._tqdm_pbar.update(1)

        def _vals_to_cols_title(self, content: Sequence[Any]) -> str:
            if _IMPORTED_COLORAMA:
                assert len(content) == len(self._cols_size)
                prefix = "" + colorama.Style.BRIGHT + colorama.Back.RED
                suffix = "" + colorama.Style.RESET_ALL
                str_c: MutableSequence[str] = []
                for i, c in enumerate(content):
                    str_c.append(("{0:<" + str(self._cols_size[i]) + "}").format(f"{c}"))
                return prefix + "\t".join(str_c) + suffix
            else:
                return super()._vals_to_cols_title(content)

except ImportError:
    class TQDMAlgorithmLogger(AlgorithmLogger): # type: ignore
        def __init__(self, *args, **kwargs):
            warnings.warn(f"`TQDMAlgorithmLogger` class needs the 'tqdm' package to be installed and importable. Using the base `AlgorithmLogger` instead.")
            super().__init__(*args, **kwargs)
            #raise NotImplementedError("`TQDMAlgorithmLogger` needs the 'tqdm' package to be installed and importable.")


# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""Collection of test functions and benchmarks for QD algorithms"""

from .base import *
from .artificial_landscapes import *

# MODELINE "{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""A collection of functions to plot containers using Matplotlib."""


import os
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from functools import reduce
from operator import mul

from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, Mapping, MutableMapping, overload

from qdpy.utils import is_iterable
from qdpy import containers
from qdpy import algorithms


########### Plots ########### {{{1




# TODO refactor name, etc
def plotGridSubplots(data, outputFilename, cmap, featuresBounds=((0., 1.), (0., 1.), (0., 1.), (0., 1.)), fitnessBounds=(0., 1.), drawCbar = True, xlabel = "", ylabel = "", cBarLabel = "", nbBins = None, nbTicks = None, binSizeInInches = 0.30):
    """TODO"""
    # Verify data dimension is supported by this funtion
    if len(data.shape) > 4:
        raise ValueError("plotGridSubplots only supports up to 4 dimensions.")
    elif len(data.shape) <= 2:
        plotGrid(data, outputFilename, cmap, featuresBounds=featuresBounds, fitnessBounds=fitnessBounds, drawCbar=drawCbar, xlabel=xlabel, ylabel=ylabel, cBarLabel=cBarLabel, nbBins=nbBins, nbTicks=nbTicks)
        return

    # Verify dimension is even
    if len(data.shape) % 2 == 1:
        data = data.reshape((data.shape[0], 1) + data.shape[1:])
        featuresBounds = (featuresBounds[0], (0., 0.)) + tuple(featuresBounds[1:]) 
        if nbBins != None:
            nbBins = (nbBins[0], 1) + nbBins[1:]
    if not nbBins:
        nbBins = data.shape

    #data[0,:,:,:] = np.linspace(0., 1., nbBins[1] * nbBins[2] * nbBins[3]).reshape((nbBins[1], nbBins[2], nbBins[3]))

    # Compute figure infos from nbBins
    horizNbBins = nbBins[::2]
    horizNbBinsProd = reduce(mul, horizNbBins, 1)
    vertNbBins = nbBins[1::2]
    vertNbBinsProd = reduce(mul, vertNbBins, 1)
    totProp = horizNbBinsProd + vertNbBinsProd
    upperlevelTot = nbBins[0] + nbBins[1]

    # Determine figure size from nbBins infos
    #figsize = [2.1 + 10. * horizNbBinsProd / upperlevelTot, 1. + 10. * vertNbBinsProd / upperlevelTot]
    #if figsize[1] < 2:
    #    figsize[1] = 2.
    figsize = [2.1 + horizNbBinsProd * binSizeInInches, 1. + vertNbBinsProd * binSizeInInches]

    # Create figure
    fig, axes = plt.subplots(nrows=nbBins[1], ncols=nbBins[0], figsize=figsize)

    # Create subplots
    for x in range(nbBins[0]):
        for y in range(nbBins[1]):
            ax = plt.subplot(nbBins[1], nbBins[0], (nbBins[1] - y - 1) * nbBins[0] + x + 1)
            #ax = axes[x,y]
            cax = drawGridInAx(data[x, y, 0:nbBins[2], 0:nbBins[3]], ax, cmap=cmap, featuresBounds=featuresBounds[-2:], fitnessBounds=fitnessBounds[-2:], aspect="equal", xlabel=xlabel, ylabel=ylabel, nbBins=(nbBins[2], nbBins[3]), nbTicks=nbTicks)

    plt.tight_layout()
    if drawCbar:
        fig.subplots_adjust(right=0.85, wspace=0.40)
        #cbarAx = fig.add_axes([0.90, 0.15, 0.01, 0.7])
        if figsize[0] < 4.:
            cbarAx = fig.add_axes([0.75, 0.15, 0.02, 0.7])
        elif figsize[0] < 6.:
            cbarAx = fig.add_axes([0.80, 0.15, 0.02, 0.7])
        elif figsize[0] < 10.:
            cbarAx = fig.add_axes([0.85, 0.15, 0.02, 0.7])
        else:
            cbarAx = fig.add_axes([0.90, 0.15, 0.02, 0.7])
        cbar = fig.colorbar(cax, cax=cbarAx, format="%.2f")
        cbar.ax.tick_params(labelsize=20)
        cbar.ax.set_ylabel(cBarLabel, fontsize=22)

    fig.savefig(outputFilename)





# TODO refactor name, etc
def drawGridInAx(data, ax, cmap, featuresBounds, fitnessBounds, aspect="equal", xlabel = "", ylabel = "", nbBins=None, nbTicks = 5):
    # Determine bounds
    vmin = fitnessBounds[0]
    if np.isnan(vmin) or np.isinf(vmin):
        vmin = np.nanmin(data)
    vmax = fitnessBounds[1]
    if np.isnan(vmax) or np.isinf(vmax):
        vmax = np.nanmax(data)

    # Draw grid
    cax = ax.imshow(data.T, interpolation="none", cmap=cmap, vmin=vmin, vmax=vmax, aspect=aspect)
    ax.invert_yaxis()

    # Define the number of ticks on x,y axis
    if is_iterable(nbTicks):
        if len(nbTicks) != 2:
            raise ValueError("nbTicks can be None, an Integer or a Sequence of size 2.")
        nbTicksX, nbTicksY = nbTicks
    elif nbTicks == None:
        nbTicksX = round(pow(nbBins[0], 1./2.))
        nbTicksX = nbTicksX if nbTicksX % 2 == 0 else nbTicksX + 1
        nbTicksY = round(pow(nbBins[1], 1./2.))
        nbTicksY = nbTicksY if nbTicksY % 2 == 0 else nbTicksY + 1
    else:
        if nbBins[0] > nbBins[1]:
            nbTicksX = nbTicks
            nbTicksY = int(nbTicksX * nbBins[1] / nbBins[0])
        elif nbBins[1] > nbBins[0]:
            nbTicksY = nbTicks
            nbTicksX = int(nbTicksY * nbBins[0] / nbBins[1])
        else:
            nbTicksX = nbTicksY = nbTicks
        # Verify than the number of ticks is valid
        if nbTicksX > nbBins[0] or nbTicksX < 1:
            nbTicksX = min(nbBins[0], nbTicks)
        if nbTicksY > nbBins[1] or nbTicksY < 1:
            nbTicksY = min(nbBins[1], nbTicks)

    # Set ticks
    ax.xaxis.set_tick_params(which='major', left=True, bottom=True, top=False, right=False)
    ax.yaxis.set_tick_params(which='major', left=True, bottom=True, top=False, right=False)
    if len(featuresBounds) > 1:
        xticks = list(np.arange(0, data.shape[0] + 1, data.shape[0] / nbTicksX))
        yticks = list(np.arange(0, data.shape[1] + 1, data.shape[1] / nbTicksY))
        deltaFeature0 = featuresBounds[0][1] - featuresBounds[0][0]
        ax.set_xticklabels([round(float(x / float(data.shape[0]) * deltaFeature0 + featuresBounds[0][0]), 2) for x in xticks], fontsize=22)
        deltaFeature1 = featuresBounds[1][1] - featuresBounds[1][0]
        ax.set_yticklabels([round(float(y / float(data.shape[1]) * deltaFeature1 + featuresBounds[1][0]), 2) for y in yticks], fontsize=22)
        plt.xticks(xticks, rotation='vertical')
    else:
        yticks = list(np.arange(0, data.shape[1] + 1, data.shape[1] / nbTicksY))
        deltaFeature0 = featuresBounds[0][1] - featuresBounds[0][0]
        ax.set_yticklabels([round(float(y / float(data.shape[1]) * deltaFeature0 + featuresBounds[0][0]), 2) for y in yticks], fontsize=22)
        plt.xticks([])
        ax.set_xticklabels([])
    if nbBins[1] == 1:
        yticks = []
    plt.yticks(yticks)

    # Draw grid
    ax.xaxis.set_tick_params(which='minor', direction="in", left=False, bottom=False, top=False, right=False)
    ax.yaxis.set_tick_params(which='minor', direction="in", left=False, bottom=False, top=False, right=False)
    ax.set_xticks(np.arange(-.5, data.shape[0], 1), minor=True)
    ax.set_yticks(np.arange(-.5, data.shape[1], 1), minor=True)
    #ax.grid(which='minor', color=(0.8,0.8,0.8,0.5), linestyle='-', linewidth=0.1)

    ax.set_xlabel(xlabel, fontsize=25)
    ax.set_ylabel(ylabel, fontsize=25)
    ax.autoscale_view()
    return cax



# TODO refactor name, etc
def plotGrid(data, outputFilename, cmap, featuresBounds=[(0., 1.), (0., 1.)], fitnessBounds=(0., 1.), drawCbar = True, xlabel = "", ylabel = "", cBarLabel = "", nbBins = None, nbTicks = None):
    if len(data.shape) == 1:
        data = data.reshape((data.shape[0], 1))
        featuresBounds = tuple(featuresBounds) + ((0., 0.),)
        if nbBins != None:
            nbBins = nbBins + (1,)
    elif len(data.shape) > 2:
        raise ValueError("plotGrid only supports 1 ou 2-dimensional data.")
    if not nbBins:
        nbBins = data.shape

    figsize = [2.1 + 10. * nbBins[0] / (nbBins[0] + nbBins[1]), 1. + 10. * nbBins[1] / (nbBins[0] + nbBins[1])]
    aspect = "equal"
    if figsize[1] < 2:
        figsize[1] = 2.
        aspect = "auto"

    fig, ax = plt.subplots(figsize=figsize)
    cax = drawGridInAx(data, ax, cmap=cmap, featuresBounds=featuresBounds, fitnessBounds=fitnessBounds, aspect=aspect, xlabel=xlabel, ylabel=ylabel, nbBins=nbBins, nbTicks=nbTicks)

    if drawCbar:
        divider = make_axes_locatable(ax)
        #cax2 = divider.append_axes("right", size="5%", pad=0.15)
        cax2 = divider.append_axes("right", size=0.5, pad=0.15)
        cbar = fig.colorbar(cax, cax=cax2, format="%.2f")
        cbar.ax.tick_params(labelsize=22)
        cbar.ax.set_ylabel(cBarLabel, fontsize=24)

    plt.tight_layout()
    fig.savefig(outputFilename)




def plot_evals(logger, output_filename, key="max", ylim=None, ylabel=None, figsize=(4., 4.)):
    """Plot the evolution of statistical parameter `key` according to the number of evaluations.  """
    # Retrieve evaluation data for key `key`
    if isinstance(logger, algorithms.AlgorithmLogger):
        data = logger.evals[key]
    else:
        data = logger
    #print(data)

    fig, ax = plt.subplots(figsize=figsize)
    #ax = fig.add_subplot(111)
    #fig.subplots_adjust(bottom=0.3)

    x = np.arange(len(data))

    # Normalise data
    tmp_shape = 1
    for i in x:
        if is_iterable(data[i]):
            tmp_shape = max(tmp_shape, len(data[i]))
    data_shape = (len(data), tmp_shape)
    normalised_data = np.zeros(data_shape)
    for i in x:
        if is_iterable(data[i]):
            normalised_data[i] = np.array(data[i])
        else:
            normalised_data[i, 0] = data[i]
    ax.plot(x, normalised_data, color='k')

    if ylim is not None:
        ax.set_ylim(ylim)

    #x = np.arange(0, nb_iterations+1, 25)
    #if len(x) > 4:
    #    x = x[::2]
    #tsplot(ax, data, color='k')
    #ax.set_ylim([0., 1.])
    plt.xlabel("Evaluations", fontsize=14)
    #plt.xticks(x, fontsize=18)
    #ax.set_xticklabels([str(i * args.nbEvalsPerIt) for i in x])
    ylabel_ = ylabel if ylabel is not None else key
    plt.ylabel(ylabel_, fontsize=14)
    #plt.yticks(fontsize=18)
    #sns.despine()
    #plt.tight_layout(rect=[0, 0, 1.0, 0.95])
    plt.tight_layout()
    fig.savefig(output_filename)
    plt.close(fig)



def plot_iterations(logger, output_filename, key="max", ylim=None, ylabel=None, figsize=(4., 4.)):
    """Plot the evolution of statistical parameter `key` according to the number of iterations."""
    # Retrieve evaluation data for key `key`
    if isinstance(logger, algorithms.AlgorithmLogger):
        data = logger.iterations[key]
    else:
        data = logger

    fig, ax = plt.subplots(figsize=figsize)
    x = np.arange(len(data))

    # Normalise data
    tmp_shape = 1
    for i in x:
        if is_iterable(data[i]):
            tmp_shape = max(tmp_shape, len(data[i]))
    data_shape = (len(data), tmp_shape)
    normalised_data = np.zeros(data_shape)
    for i in x:
        if is_iterable(data[i]):
            normalised_data[i] = np.array(data[i])
        else:
            normalised_data[i, 0] = data[i]
    ax.plot(x, normalised_data, color='k')

    if ylim is not None:
        ax.set_ylim(ylim)

    plt.xlabel("Iterations", fontsize=14)
    ylabel_ = ylabel if ylabel is not None else key
    plt.ylabel(ylabel_, fontsize=14)
    plt.tight_layout()
    fig.savefig(output_filename)
    plt.close(fig)


def default_plots_grid(logger, output_dir=None, to_grid_parameters={}, fitness_domain=None):
    """Make all default plots for algorithms using grid-based or grid-convertible containers."""
    container = logger.algorithms[0].container # XXX
    if output_dir is None:
        output_dir = logger.log_base_path

    plot_evals(logger, os.path.join(output_dir, "./evals_fitnessmax0.pdf"), "max0", ylabel="Fitness")
    ylim_contsize = (0, len(container)) if np.isinf(container.capacity) else (0, container.capacity)
    plot_evals(logger, os.path.join(output_dir, "./evals_contsize.pdf"), "cont_size", ylim=ylim_contsize, ylabel="Container size")
    plot_iterations(logger, os.path.join(output_dir, "./iterations_nbupdated.pdf"), "nb_updated", ylabel="Number of updated bins")

    if isinstance(container, containers.Grid):
        grid = container
    else:
        if 'shape' not in to_grid_parameters:
            to_grid_parameters['shape'] = (32,) * len(container.features_domain)
        grid = container.to_grid(**to_grid_parameters)

    plot_path = os.path.join(output_dir, "performancesGrid.pdf")
    cmap_perf = "inferno" if logger.algorithms[0].optimisation_task == "maximisation" else "inferno_r"
    fitness_domain = grid.fitness_domain if fitness_domain is None else fitness_domain
    plotGridSubplots(grid.quality_array[... ,0], plot_path, plt.get_cmap(cmap_perf), grid.features_domain, fitness_domain[0], nbTicks=None)

    plot_path = os.path.join(output_dir, "activityGrid.pdf")
    max_activity = np.max(grid.activity_per_bin)
    plotGridSubplots(grid.activity_per_bin, plot_path, plt.get_cmap("Reds", max_activity), grid.features_domain, [0, max_activity], nbTicks=None)



# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""All the base classes and protocols describing QD algorithms (optimisers). All of these algorithms perform optimisation on a defined objective functions, and use a provided container to store the evaluated solutions. The base abstract class of algorithms is `QDAlgorithm`, itself implementing the `QDAlgorithmLike` Protocol. Most QD algorithms will subclass QDAlgorithm directly, but in some special cases, algorithms might need to subclass QDAlgorithmLike instead (e.g. sequences `Sq` of QD algorithms).

All algorithms must be registered to the default qdpy.base.Registry `qdpy.base.registry` if they are to be easily accessed or created by factories. It can be done by decorating QD algorithms classes with `@registry.register`. """

__all__ = ["partial", "QDAlgorithmLike", "QDAlgorithm", "Lambda", "AlgWithBatchSuggestions", "Sq"]

import abc
import math
import time
from timeit import default_timer as timer
from inspect import signature
from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, Mapping, MutableMapping, overload
from typing_extensions import runtime, Protocol
import warnings
import numpy as np
import copy
import traceback


from qdpy.utils import *
from qdpy.base import *
from qdpy.containers import *
from qdpy import tools

import functools
partial = functools.partial

default_algorithm_logger: Any

def _evalWrapper(eval_id: int, fn: Callable, *args, **kwargs) -> Tuple[int, float, Any, Any]:
    """Wrapper around an evaluation function. It catches any exceptions raised by the evaluation function (as exceptions might be lost depending on which kind of parallelism scheme is applied to launch evaluations). It also measures the time elapsed by the evaluation function computation."""
    start_time = timer()
    exc = None
    fitness = None
    features = None
    try:
        res = fn(*args, **kwargs)
    except Exception as e:
        print(f"Exception during evaluation: {e}")
        traceback.print_exc()
        exc = e
    elapsed = timer() - start_time
    return eval_id, elapsed, res, exc


def _severalEvalsWrapper(eval_id: int, fn: Callable, *args, **kwargs) -> Tuple[int, float, Any, Any]:
    """Wrapper around several evaluations of an objective function. It catches any exceptions raised by the evaluation function (as exceptions might be lost depending on which kind of parallelism scheme is applied to launch evaluations). It also measures the time elapsed by the evaluation function computation."""
    start_time = timer()
    exc = None
    fitness = None
    features = None
    try:
        res = fn(*args, **kwargs)
    except Exception as e:
        print(f"Exception during evaluation: {e}")
        traceback.print_exc()
        exc = e
    elapsed = timer() - start_time
    return eval_id, elapsed, res, exc


class _CallbacksManager(Copyable):
    """Manager of callback functions that are called when specific events of the `QDAlgorithm` class are triggered.
    It is possible to register a callback function using the `self.add_callback` method. The callback functions
    attached to a specific event are retrieved through the `self.get` method.

    We regroup callback function handling in this private class to ease the serialisation of QDAlgorithm and its sub-classes.
    Namely, the callback functions provided by the user may not be pickleable.  To prevent them to be taken into account during serialisation, we declare a __getstate__ method returning the __dict__ dictionary of object attributes, minus the registered callback function objects. This __getstate__ is not inherited by sub-classes, which make it unadapted to be specified directly in the QDAlgorithm class, as it would force all sub-classes of QDAlgorithm to redefine the __getstate__ method to be pickleable."""
    callbacks: MutableMapping[str, MutableSequence[Callable]]

    def __init__(self):
        self.callbacks = {}
        pass

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['callbacks']
        return odict

    def get(self, event: str) -> Sequence[Callable]:
        """Return the Sequence of functions that can be triggered when `event` happens. 

        Parameters
        ----------
        :param event: str
            The event triggering the returned callback functions.
        """
        return self.callbacks.get(event, [])

    def add_callback(self, event: str, fn: Callable) -> None:
        """Attach a callable `fn` to this CallbacksManager. `fn` will be called each time the event specified in `event` occurs.
        
        Parameters
        ----------
        :param event: str
            The event triggering the callback.
        :param fn: Callable
            The callback function called when the event specified in `event` occurs.
        """
        self.callbacks.setdefault(event, []).append(fn)


@runtime
class QDAlgorithmLike(Protocol):
    """Protocol (Python interfaces) describing QD algorithms. All classes describing qdpy algorithms must implement this protocol -- or alternatively the `qdpy.algorithms.QDAlgorithm` abstract class, which already implements `qdpy.algorithms.QDAlgorithmLike`.

    Qdpy algorithms are based around the ask-and-tell paradigm, introduced by [Collete2010].
    TODO description ask-and-tell

    TODO description functions

    References
    ----------
    [Collete2010] TODO
    """
    name: str
    container: Container
    dimension: Optional[int]
    evaluate: Optional[Callable]
    budget: int

    def ask(self) -> IndividualLike: ...

    def tell(self, individual: IndividualLike, fitness: Optional[Any] = None,
            features: Optional[FeaturesLike] = None, elapsed: Optional[float] = None) -> None: ...

    def tell_container_entries(self, cont: Optional[Container] = None) -> None: ...

    def best(self) -> IndividualLike: ...

    def optimise(self, evaluate: Callable, budget: Optional[int] = None, 
            batch_mode: bool = True, executor: Optional[ExecutorLike] = None,
            send_several_suggestions_to_fn: bool = False) -> IndividualLike: ...

    def add_callback(self, event: str, fn: Callable) -> None: ...

    @property
    def nb_suggestions(self) -> int: ...

    @property
    def nb_suggestions_in_iteration(self) -> int: ...

    @property
    def nb_evaluations(self) -> int: ...

    @property
    def nb_evaluations_in_iteration(self) -> int: ...

    @property
    def nb_updated(self) -> int: ...

    @property
    def nb_updated_in_iteration(self) -> int: ...

    @property
    def batch_size(self) -> int: ...

    @property
    def nb_max_iter(self) -> int: ...

    @property
    def current_iter(self) -> int: ...


# TODO say inspiration from nevergrad
class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, CreatableFromConfig):
    """TODO"""

    #ind_domain: Optional[DomainLike]
    #executor: Optional[ExecutorLike]

    name: str
    container: Container
    dimension: Optional[int]
    evaluate: Optional[Callable] # XXX complete type
    budget: int
    _base_ind_gen: Optional[Generator[IndividualLike, None, None]]
    _nb_objectives: int
    _optimisation_task: Union[str, Sequence[float]]
    #_elapsed: float = 0.
    _fitness_weights: Sequence
    _nb_suggestions: int
    _nb_suggestions_in_iteration: int
    _nb_evaluations: int
    _nb_evaluations_in_iteration: int
    _nb_updated: int
    _nb_updated_in_iteration: int
    _batch_size: Optional[int]
    _callbacks: _CallbacksManager


    def __init__(self, container: Container, budget: int, batch_size: Optional[int] = None,
            dimension: Optional[int] = None, nb_objectives: Optional[int] = None,
            optimisation_task: Union[str, Sequence[float]] = "minimisation",
            base_ind_gen: Optional[Generator[IndividualLike, None, None]] = None,
            tell_container_at_init: Union[bool, str] = False, # XXX
            add_default_logger: bool = False, name: Optional[str] = None,
            **kwargs):
        self.container = container
        self.dimension = dimension
        self.budget = budget
        self._batch_size = batch_size
        self._base_ind_gen = base_ind_gen
        self.name = name if name is not None else f"{self.__class__.__name__}-{id(self)}"

        if nb_objectives is not None:
            self._nb_objectives = nb_objectives
        else:
            if self.container.fitness_domain is None or len(self.container.fitness_domain) == 0:
                #raise ValueError("You must either specify `nb_objectives` or providing as `container` a Container with a declared fitness domain.")
                warnings.warn(f"Unspecified number of objectives ! When creating a QDAlgorithm class, you must either specify `nb_objectives` or providing as `container` a Container with a declared fitness domain. Defaulting to only 1 objective.")
                self._nb_objectives = 1
            else:
                self._nb_objectives = len(self.container.fitness_domain)

        self._nb_suggestions = 0
        self._nb_suggestions_in_iteration = 0
        self._nb_evaluations = 0
        self._nb_evaluations_in_iteration = 0
        self._nb_updated = 0
        self._nb_updated_in_iteration = 0
        self._callbacks = _CallbacksManager()
        self.optimisation_task = optimisation_task
        if add_default_logger:
            default_algorithm_logger.monitor(self)

        # If needed, tell container entries
        tell_container_depot: bool = False
        tell_container: bool = False
        if isinstance(tell_container_at_init, bool):
            tell_container = tell_container_at_init
        else:
            if tell_container_at_init.lower() == "yes":
                tell_container = True
            elif tell_container_at_init.lower() == "depot":
                tell_container_depot = True
        if tell_container_depot:
            if self.container.depot is None:
                raise ValueError("`tell_container_depot` can only be set to 'depot' if `container` possesses a depot.")
            for e in self.container.depot:
                is_in_container = e in self.container
                self._internal_tell(e, is_in_container)
        elif tell_container:
            for e in self.container:
                self._internal_tell(e, True)


    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['_base_ind_gen']
        return odict

    @property
    def optimisation_task(self) -> Union[str, Sequence[float]]:
        return self._optimisation_task

    @optimisation_task.setter
    def optimisation_task(self, optimisation_task: Union[str, Sequence[float]]) -> None:
        """Update the kind of optimisation task performed by this `QDAlgorithm`.
        
        Parameters
        ----------
        :param optimisation_task: Union[str, Sequence[float]]
            The kind of optimisation task to perform for all objectives of the fitness functions.
            If `optimisation_task` is a string, it can take the following values:
                - for minimising all fitness objectives: 'minimisation', 'minimization', 'minimise', 'minimize', 'min'
                - for maximising all fitness objectives: 'maximisation', 'maximization', 'maximise', 'maximize', 'max'
            If `optimisation_task` is a Sequence of floats, its length corresponds to the number of objectives, and each
            entry is a weight to apply to each objective score.
            Maximised objectives have a corresponding weight > 0 while minimised objectives have weights < 0.
            The length of this Sequence must be same as the length of the `self.container.fitness_domain` attribute.
        """
        if isinstance(optimisation_task, str) and \
                any((optimisation_task == a for a in ["minimisation", "minimization", "minimise", "minimize", "min"])):
            self._optimisation_task = "minimisation"
            self._fitness_weights = (-1.,) * self._nb_objectives
        elif isinstance(optimisation_task, str) and \
                any((optimisation_task == a for a in ["maximisation", "maximization", "maximise", "maximize", "max"])):
            self._optimisation_task = "maximisation"
            self._fitness_weights = (1.,) * self._nb_objectives
        elif isinstance(optimisation_task, Sequence) and not isinstance(optimisation_task, str) and len(optimisation_task) > 0:
            self._optimisation_task = optimisation_task
            self._fitness_weights = optimisation_task
        else:
            raise ValueError("`optimisation_task` must be either a tuple of fitness weights, or a string equals to either 'minimisation' or 'maximisation'.")


    @property
    def nb_suggestions(self) -> int:
        return self._nb_suggestions

    @property
    def nb_suggestions_in_iteration(self) -> int:
        return self._nb_suggestions_in_iteration

    @property
    def nb_evaluations(self) -> int:
        return self._nb_evaluations

    @property
    def nb_evaluations_in_iteration(self) -> int:
        return self._nb_evaluations_in_iteration

    @property
    def nb_updated(self) -> int:
        return self._nb_updated

    @property
    def nb_updated_in_iteration(self) -> int:
        return self._nb_updated_in_iteration

    @property
    def batch_size(self) -> int:
        return self._batch_size if self._batch_size is not None else self.budget

    @property
    def nb_max_iter(self) -> int:
        return math.ceil(self.budget / self.batch_size)

    @property
    def current_iter(self) -> int:
        return math.floor(self._nb_evaluations / self.batch_size)


    def ask(self) -> IndividualLike:
        """Ask for a suggestion of an individual to evaluate.

        Returns
        -------
        out: IndividualLike
            The suggested individual.
        """
        # Create base individual
        if self._base_ind_gen is None:
            base_ind: IndividualLike = Individual()
        else:
            base_ind = next(self._base_ind_gen)
        # Call _internal_ask
        suggestion = self._internal_ask(base_ind)
        # Update individual name and fitness weights
        if suggestion.name == "":
            suggestion.name = f"{int(round(time.time() * 1000))}-{self._nb_suggestions}"
        suggestion.fitness.weights = self._fitness_weights
        self._nb_suggestions += 1
        self._nb_suggestions_in_iteration += 1
        # Call callback functions
        for fn in self._callbacks.get("ask"):
            fn(self, suggestion)
        return suggestion


    @abc.abstractmethod
    def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
        """Internal ask method. To be implemented by subclasses.

        Parameters
        ----------
        :param base_ind: IndividualLike
            The base (empty) individual object. The `_internal_ask` method must change
            this object, by filling out the genome and providing other information.
        
        Returns
        -------
        out: IndividualLike
            The suggested individual.
        """
        raise NotImplementedError("Method `_internal_ask` must be implemented by QDAlgorithm subclasses.")


    def tell(self, individual: IndividualLike, fitness: Optional[Any] = None,
            features: Optional[FeaturesLike] = None, elapsed: Optional[float] = None) -> None:
        """Give the algorithm an evaluated `individual` (with a valid fitness function)
        and store it into the container.

        Parameters
        ----------
        :param individual: IndividualLike
            The `individual` provided to the algorithm.
        :param fitness: Optional[FitnessLike]
            If provided, set `individual.fitness` to `fitness`.
        :param features: Optional[FeaturesLike]
            If provided, set `individual.features` to `features`.
        :param elapsed: Optional[float]
            If provided, set `individual.elapsed` to `elapsed`.
            It corresponds to the time (in seconds) elapsed during evaluation.
        """
        ind = copy.deepcopy(individual)
        if fitness is not None:
            if isinstance(fitness, FitnessLike):
                ind.fitness = fitness
            else:
                ind.fitness.values = fitness
        if len(ind.fitness.values) != self._nb_objectives:
            raise ValueError(f"`individual` must possess a `fitness` attribute of dimension {self._nb_objectives} (specified through parameter `nb_objectives` when creating a QDAlgorithm sub-class), instead of {len(individual.fitness.values)}.")
        if features is not None:
            ind.features = features
        if elapsed is not None:
            ind.elapsed = elapsed
        added: bool = self.container.update([ind]) == 1
        self._tell_after_container_update(ind, added)


    def _tell_after_container_update(self, individual: IndividualLike, added: bool) -> None:
        if added:
            self._nb_updated += 1
            self._nb_updated_in_iteration += 1
        self._internal_tell(individual, added)
        self._nb_evaluations += 1
        self._nb_evaluations_in_iteration += 1
        # Call callback functions
        for fn in self._callbacks.get("tell"):
            fn(self, individual)


    def _internal_tell(self, individual: IndividualLike, added_to_container: bool) -> None:
        """Internal tell method. To be overrode by subclasses, or left empty.
        
        Parameters
        ----------
        :param individual: IndividualLike
            The `individual` provided to the algorithm.
        :param added_to_container: bool
            True if `individual` was successfully added to the container.
        """
        pass


    def tell_container_entries(self, cont: Optional[Container] = None) -> None:
        """Tell this algorithm all entries from a container. By default, use
        `self.container` as container of interest, but if `cont` is specified,
        it will it instead.

        Parameters
        ----------
        :param cont: Optional[Container] = None
            If None, this method will tell all entries from the local container `self.container`.
            If set, this method will tell all entries from container `cont`.
        """
        if cont is None:
            cont = self.container
        for ind in cont:
            self._internal_tell(ind, True)


    def best(self) -> IndividualLike:
        """Returns the individual with the best quality.
        Raise a RuntimeError if there are no individual in the container."""
        if len(self.container) == 0:
            raise RuntimeError("`self.container` contains no individual !")
        return self._internal_best()


    def _internal_best(self) -> IndividualLike:
        """Returns the individual with the best quality. To be overrode by subclasses.
        By default, returns the best individual of `self.container`."""
        return self.container.best # Assume `self.container` is not empty



    def _verify_if_finished_iteration(self, batch_start_time: float) -> bool:
        """TODO"""
        # Verify if we finished an iteration
        if self._nb_evaluations % self.batch_size != 0:
            return False
        batch_elapsed: float = timer() - batch_start_time
        # Call callback functions
        for fn in self._callbacks.get("iteration"):
            fn(self, batch_elapsed)
        self._nb_suggestions_in_iteration = 0
        self._nb_evaluations_in_iteration = 0
        self._nb_updated_in_iteration = 0
        return True


    def optimise(self, evaluate: Callable, budget: Optional[int] = None, 
            batch_mode: bool = True, executor: Optional[ExecutorLike] = None,
            send_several_suggestions_to_fn: bool = False) -> IndividualLike:
        """TODO"""
        optimisation_start_time: float = timer()
        if budget is not None:
            self.budget = budget
        if evaluate is not None:
            self.evaluate = evaluate
        if self.budget is None:
            raise ValueError("`budget` must be provided.")
        _executor: ExecutorLike = executor if executor is not None else SequentialExecutor()

        # Call callback functions
        for fn in self._callbacks.get("started_optimisation"):
            fn(self)

        def optimisation_loop(remaining_evals: int):
            batch_start_time: float = timer()
            futures: MutableMapping[int, FutureLike] = {}
            individuals: MutableMapping[int, Union[IndividualLike, Sequence[IndividualLike]]] = {}
            while remaining_evals > 0 or len(futures) > 0:
                nb_suggestions = min(remaining_evals, self.batch_size - len(futures))

                if send_several_suggestions_to_fn:
                    # Launch evals on suggestions
                    eval_id: int = remaining_evals
                    inds: Sequence[IndividualLike] = [self.ask() for _ in range(nb_suggestions)]
                    individuals[eval_id] = inds
                    futures[eval_id] = _executor.submit(_severalEvalsWrapper, eval_id, self.evaluate, inds)
                    remaining_evals -= len(inds)

                    # Wait for next completed future
                    f = generic_as_completed(list(futures.values()))
                    ind_id, ind_elapsed, ind_res, ind_exc = f.result()
                    if ind_exc is None:
                        inds = individuals[ind_id] # type: ignore
                        for i in range(len(inds)):
                            self.tell(inds[i], fitness=ind_res[i][0], features=ind_res[i][1], elapsed=ind_elapsed)
                    else:
                        warnings.warn(f"Individual evaluation raised the following exception: {ind_exc} !")
                        self._nb_evaluations += len(ind_res)

                else:
                    # Launch evals on suggestions
                    for _ in range(nb_suggestions):
                        eval_id = remaining_evals
                        ind: IndividualLike = self.ask()
                        individuals[eval_id] = ind
                        futures[eval_id] = _executor.submit(_evalWrapper, eval_id, self.evaluate, ind)
                        remaining_evals -= 1

                    # Wait for next completed future
                    f = generic_as_completed(list(futures.values()))
                    #ind_id, ind_elapsed, ind_fitness, ind_features, ind_exc = f.result()
                    ind_id, ind_elapsed, ind_res, ind_exc = f.result()
                    if ind_exc is None:
                        if isinstance(ind_res, IndividualLike):
                            self.tell(ind_res, elapsed=ind_elapsed)
                        else:
                            ind_fitness, ind_features = ind_res
                            ind = individuals[ind_id] # type: ignore
                            self.tell(ind, fitness=ind_fitness, features=ind_features, elapsed=ind_elapsed)
                    else:
                        warnings.warn(f"Individual evaluation raised the following exception: {ind_exc} !")
                        self._nb_evaluations += 1

                # Clean up
                del futures[ind_id]
                del individuals[ind_id]
                # Verify if we finished an iteration
                if self._verify_if_finished_iteration(batch_start_time):
                    batch_start_time = timer()

        if batch_mode:
            remaining_evals: int = self.budget
            for _ in range(self.nb_max_iter):
                budget_iteration: int = min(remaining_evals, self.batch_size)
                optimisation_loop(budget_iteration)
                remaining_evals -= self.batch_size
        else:
            optimisation_loop(self.budget)

        # Call callback functions
        optimisation_elapsed: float = timer() - optimisation_start_time
        for fn in self._callbacks.get("finished_optimisation"):
            fn(self, optimisation_elapsed)

        return self.best()


    def add_callback(self, event: str, fn: Callable) -> None:
        """Attach a callable `fn` to the optimiser. `fn` will be called each time the event specified in `event` occurs.
        
        Parameters
        ----------
        :param event: str
            The event triggering the callback. Can be either 'ask', 'tell', 'iteration' (when an iteration finishes)
            'started_optimisation', or 'finished_optimisation'.
        :param fn: Callable
            The callback function called when the event specified in `event` occurs.
        """
        _lst_events: Sequence[str] = ['ask', 'tell', 'iteration', 'started_optimisation', 'finished_optimisation']
        if not event in _lst_events:
            raise ValueError(f"`event` can be either {','.join(_lst_events)}.")
        #self._callbacks.setdefault(event, []).append(fn)
        self._callbacks.add_callback(event, fn)



@registry.register
class Lambda(QDAlgorithm):
    """TODO"""
    _ask_fn: Union[Callable[[], IndividualLike], Callable[[IndividualLike], IndividualLike]]
    _tell_fn: Optional[Callable]

    def __getstate__(self):
        odict = super().__getstate__()
        del odict['_ask_fn']
        del odict['_tell_fn']
        return odict

    def __init__(self, container: Container, budget: int,
            ask: Union[Callable[[], IndividualLike], Callable[[IndividualLike], IndividualLike]],
            tell: Optional[Callable] = None, **kwargs):
        super().__init__(container, budget, **kwargs)
        self._ask_fn = ask
        self._tell_fn = tell

    def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
        """Internal ask method. Call the function specified by the `ask` argument of the constructor.
        This function can either take no argument (and must return a Sequence corresponding to the genome)
        or take a base (empty) `qdpy.base.IndividualLike` as sole argument
        (and must return a `qdpy.base.IndividualLike` with a filled out genome)."""
        sig = signature(self._ask_fn)
        if len(sig.parameters) == 0:
            base_ind[:] = self._ask_fn() # type: ignore
            return base_ind
        else:
            return self._ask_fn(base_ind) # type: ignore

    def _internal_tell(self, individual: IndividualLike, added_to_container: bool) -> None:
        if self._tell_fn is not None:
            self._tell_fn(individual, added_to_container)




class AlgWithBatchSuggestions(QDAlgorithm):
    """An abstract QDAlgorithm class that generates all of its suggestions batch-by-batch instead of at each call of `self.ask`.
    The `self._generate_batch_suggestions` method is called at the beginning of each batch to compute all suggestions
    for the current batch.
    """

    _batch_inds: List[IndividualLike]

    def __init__(self, container: Container, budget: int, **kwargs):
        super().__init__(container, budget, **kwargs)
        self._batch_inds = []

    def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
        if len(self._batch_inds) == 0:
            self._generate_batch_suggestions()
            #assert len(self._batch_inds) == self._batch_size, f"`self._generate_batch_suggestions` must populate `self._batch_inds` with `self.batch_size` IndividualLike objects."
        base_ind[:] = self._batch_inds.pop()
        return base_ind

    @abc.abstractmethod
    def _generate_batch_suggestions(self) -> None:
        """Compute all suggestions for the current batch and store them in `self._batch_inds`."""
        raise NotImplementedError("Method `_generate_batch_suggestions` must be implemented by AlgWithBatchSuggestions subclasses.")



@registry.register
class Sq(QDAlgorithmLike, Summarisable, Saveable, Copyable, CreatableFromConfig):
    """Sequence of algorithms. Useful to easily optimise several algorithms sequentially.  """
    algorithms: Sequence[QDAlgorithm]
    current_idx: int = 0
    current: QDAlgorithm
    tell_container_when_switching: Union[bool, str]

    def __init__(self, algorithms: Any, tell_container_when_switching: Union[bool, str] = True,
            name: Optional[str] = None, **kwargs):
        if len(algorithms) == 0:
            raise ValueError("`algorithms` must not be empty.")
        self.algorithms = list(algorithms)
        self.current_idx = 0
        self.current = self.algorithms[self.current_idx]
        if isinstance(tell_container_when_switching, str):
            if tell_container_when_switching not in ["True", "False", "only_best"]:
                raise ValueError("`tell_container_when_switching` must be a bool or one of these strings: 'True', 'False', 'only_best'.")
        self.tell_container_when_switching = tell_container_when_switching
        self.name = name if name is not None else f"{self.__class__.__name__}-{id(self)}"
        #self.budget = sum((algo.budget for algo in self.algorithms))
        self.switch_to(0)


    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        if "current" not in self.__dict__:
            raise AttributeError
        return getattr(self.current, attr)


    def __get_saved_state__(self) -> Mapping[str, Any]:
        """Return a dictionary containing the relevant information to save. Only include information from `self.algorithms`."""
        res = copy.deepcopy(self.current.__get_saved_state__())
        entries = {k:v for k,v in inspect.getmembers(self) if not k.startswith('_') and not inspect.ismethod(v)}
        return {**entries, **res}


    def next(self) -> None:
        """Switch to the next algorithm in Sequence `self.algorithms`, if there is one."""
        if self.current_idx < len(self.algorithms) - 1:
            self.switch_to(self.current_idx + 1)

    def switch_to(self, idx: int):
        """Switch to another algorithm of index `idx` in `self.algorithms`."""
        assert idx < len(self.algorithms), f"`idx` must be < len(self.algorithms)."
        previous = self.current
        self.current_idx = idx
        self.current = self.algorithms[idx]
        if (isinstance(self.tell_container_when_switching, bool) and self.tell_container_when_switching) or \
                self.tell_container_when_switching == "True":
            self.current.tell_container_entries()
        elif self.tell_container_when_switching == "only_best":
            try:
                self.current.tell(previous.best()) # Exception when no individual is present
            except Exception:
                pass

    def best(self) -> IndividualLike:
        return self.current.best()

    def ask(self) -> IndividualLike:
        return self.current.ask()

    def tell(self, individual: IndividualLike, fitness: Optional[Any] = None,
            features: Optional[FeaturesLike] = None, elapsed: Optional[float] = None) -> None:
        self.current.tell(individual, fitness, features, elapsed)

    def optimise(self, evaluate: Callable, budget: Optional[int] = None, 
            batch_mode: bool = True, executor: Optional[ExecutorLike] = None,
            send_several_suggestions_to_fn: bool = False) -> IndividualLike:
        for _ in range(self.current_idx, len(self.algorithms)):
            try:
                res = self.current.optimise(evaluate, budget, batch_mode, executor, send_several_suggestions_to_fn)
            except Exception as e:
                warnings.warn(f"Optimisation failed with algorithm '{self.current.name}': {e}")
                traceback.print_exc()
            self.next()
        return res

    def add_callback(self, event: str, fn: Callable) -> None:
        for a in self.algorithms:
            a.add_callback(event, fn)



# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""TODO"""

__all__ = ["qdSimple", "DEAPQDAlgorithm"]

import os
import deap.tools
import deap.algorithms
import numpy as np
from timeit import default_timer as timer
import pickle
import copy

from qdpy.containers import *


def qdSimple(init_batch, toolbox, container, batch_size, niter, cxpb = 0.0, mutpb = 1.0, stats = None, halloffame = None, verbose = False, start_time = None, iteration_callback = None):
    """The simplest QD algorithm using DEAP.
    :param init_batch: Sequence of individuals used as initial batch.
    :param toolbox: A :class:`~deap.base.Toolbox` that contains the evolution operators.
    :param batch_size: The number of individuals in a batch.
    :param niter: The number of iterations.
    :param stats: A :class:`~deap.tools.Statistics` object that is updated inplace, optional.
    :param halloffame: A :class:`~deap.tools.HallOfFame` object that will
                       contain the best individuals, optional.
    :param verbose: Whether or not to log the statistics.
    :param start_time: Starting time of the illumination process, or None to take the current time.
    :param iteration_callback: Optional callback funtion called when a new batch is generated. The callback function parameters are (iteration, batch, container, logbook).
    :returns: The final batch
    :returns: A class:`~deap.tools.Logbook` with the statistics of the
              evolution

    TODO
    """
    if start_time == None:
        start_time = timer()
    logbook = deap.tools.Logbook()
    logbook.header = ["iteration", "containerSize", "evals", "nbUpdated"] + (stats.fields if stats else []) + ["elapsed"]

    if len(init_batch) == 0:
        raise ValueError("``init_batch`` must not be empty.")

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in init_batch if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit[0]
        ind.features = fit[1]

    if len(invalid_ind) == 0:
        raise ValueError("No valid individual found !")

    # Update halloffame
    if halloffame is not None:
        halloffame.update(init_batch)

    # Store batch in container
    nb_updated = container.update(init_batch)
    if nb_updated == 0:
        raise ValueError("No individual could be added to the container !")

    # Compile stats and update logs
    record = stats.compile(container) if stats else {}
    logbook.record(iteration=0, containerSize=container.size_str(), evals=len(invalid_ind), nbUpdated=nb_updated, elapsed=timer()-start_time, **record)
    if verbose:
        print(logbook.stream)
    # Call callback function
    if iteration_callback != None:
        iteration_callback(0, init_batch, container, logbook)

    # Begin the generational process
    for i in range(1, niter + 1):
        start_time = timer()
        # Select the next batch individuals
        batch = toolbox.select(container, batch_size)

        ## Vary the pool of individuals
        offspring = deap.algorithms.varAnd(batch, toolbox, cxpb, mutpb)
        #offspring = []
        #for o in batch:
        #    newO = toolbox.clone(o)
        #    ind, = toolbox.mutate(newO)
        #    del ind.fitness.values
        #    offspring.append(ind)

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit[0]
            ind.features = fit[1]

        # Replace the current population by the offspring
        nb_updated = container.update(offspring)

        # Update the hall of fame with the generated individuals
        if halloffame is not None:
            halloffame.update(container)

        # Append the current generation statistics to the logbook
        record = stats.compile(container) if stats else {}
        logbook.record(iteration=i, containerSize=container.size_str(), evals=len(invalid_ind), nbUpdated=nb_updated, elapsed=timer()-start_time, **record)
        if verbose:
            print(logbook.stream)
        # Call callback function
        if iteration_callback != None:
            iteration_callback(i, batch, container, logbook)

    return batch, logbook



class DEAPQDAlgorithm(object):
    """TODO"""

    def __init__(self, toolbox, container = None, stats = None, halloffame = None,
            iteration_filename = "iteration-%i.p", final_filename = "final.p",
            ea_fn = qdSimple, verbose = False, results_infos = None, log_base_path = ".",
            save_period = None, iteration_callback_fn = None, **kwargs):
        self._update_params(**kwargs)
        self.toolbox = toolbox
        self.halloffame = halloffame
        self.iteration_filename = iteration_filename
        self.final_filename = final_filename
        self.ea_fn = ea_fn
        self.verbose = verbose
        self.log_base_path = log_base_path
        self.save_period = save_period
        self.iteration_callback_fn = iteration_callback_fn
        self._init_container(container)
        self._init_stats(stats)
        self._results_infos = {}
        if results_infos != None:
            self.add_results_infos(results_infos)
        self.total_elapsed = 0.


    def _init_container(self, container = None):
        """TODO"""
        if container == None:
            self.container = Container()
        else:
            self.container = container


    def _init_stats(self, stats = None):
        """TODO"""
        if stats == None:
            # Default stats
            self.stats = deap.tools.Statistics(lambda ind: ind.fitness.values)
            self.stats.register("avg", np.mean, axis=0)
            self.stats.register("std", np.std, axis=0)
            self.stats.register("min", np.min, axis=0)
            self.stats.register("max", np.max, axis=0)
        else:
            self.stats = stats


    def gen_init_batch(self, init_batch_size = None):
        """TODO"""
        if not hasattr(self, "start_time") or self.start_time == None:
            self.start_time = timer()
        if init_batch_size != None:
            self.init_batch_size = init_batch_size
        if not hasattr(self, "init_batch_size") or self.init_batch_size == None:
            raise ValueError("Please specify 'init_batch_size'")
        self.init_batch = self.toolbox.population(n = self.init_batch_size)


    def _update_params(self, **kwargs):
        """TODO"""
        for k,v in kwargs.items():
            if v != None:
                if k == "init_batch_size" or k == "batch_size" or k == "niter" or k == "start_time" or k == "save_period" or k == "iteration_filename" or k == "final_filename" or k == "log_base_path":
                    setattr(self, k, v)


    def _iteration_callback(self, iteration, batch, container, logbook):
        """TODO"""
        self.current_iteration = iteration
        self.current_batch = batch
        self.logbook = logbook
        if self.iteration_callback_fn is not None:
            self.iteration_callback_fn(iteration, batch, container, logbook)
        if self.save_period == None or self.save_period == 0:
            return
        if iteration % self.save_period == 0 and iteration != self.niter and self.iteration_filename != None and self.iteration_filename != "":
            self.save(os.path.join(self.log_base_path, self.iteration_filename % self.current_iteration))


    def run(self, init_batch = None, **kwargs):
        """TODO"""
        self._update_params(**kwargs)
        # If needed, generate the initial batch
        if init_batch == None:
            if not hasattr(self, "init_batch") or self.init_batch == None:
                self.gen_init_batch()
        else:
            self.init_batch = init_batch
        # Run the illumination process !
        batch, logbook = self.ea_fn(self.init_batch, self.toolbox, self.container, self.batch_size, self.niter, stats = self.stats, halloffame = self.halloffame, verbose = self.verbose, start_time = self.start_time, iteration_callback = self._iteration_callback)
        if self.final_filename != None and self.final_filename != "":
            self.save(os.path.join(self.log_base_path, self.final_filename))
        self.total_elapsed = timer() - self.start_time
        return self.total_elapsed


    def data_archive(self):
        """TODO"""
        results = {}
        def copy_attr(obj, names):
            for name in names:
                if hasattr(obj, name):
                    results[name] = getattr(obj, name)
        copy_attr(self, ['init_batch_size', 'batch_size', 'niter', 'container', 'current_iteration', 'current_batch', 'logbook', 'container'])
        #copy_attr(self.container, ['features_domain', 'fitness_domain', 'solutions', 'fitness', 'features', 'best_fitness', 'best_fitness_array'])
        results = {**results, **self._results_infos}
        return results


    def save(self, outputFile):
        """TODO"""
        results = self.data_archive()
        with open(outputFile, "wb") as f:
            pickle.dump(results, f)


    def add_results_infos(self, *args):
        """TODO"""
        if len(args) == 1:
            self._results_infos = {**self._results_infos, **args[0]}
        elif len(args) == 2:
            self._results_infos[args[0]] = args[1]
        else:
            raise ValueError("Please either pass a dictionary or key, value as parameter.")




# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker
